#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
#     ||          ____  _ __
#  +------+      / __ )(_) /_______________ _____  ___
#  | 0xBC |     / __  / / __/ ___/ ___/ __ `/_  / / _ \
#  +------+    / /_/ / / /_/ /__/ /  / /_/ / / /_/  __/
#   ||  ||    /_____/_/\__/\___/_/   \__,_/ /___/\___/
#
#  Copyright (C) 2011-2023 Bitcraze AB
#
#  Crazyflie Nano Quadcopter Client
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either version 2
#  of the License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.

#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
#  02110-1301, USA.

"""
An example template for a tab in the Crazyflie Client. It comes pre-configured
with the necessary QT Signals to wrap Crazyflie API callbacks and also
connects the connected/disconnected callbacks.
"""

import logging

from PyQt6 import uic
from PyQt6.QtCore import pyqtSignal
from PyQt6.QtWidgets import QMessageBox

import cfclient
from cfclient.ui.tab_toolbox import TabToolbox

__author__ = 'Bitcraze AB'
__all__ = ['ExampleTab']

logger = logging.getLogger(__name__)

example_tab_class = uic.loadUiType(cfclient.module_path + "/ui/tabs/exampleTab.ui")[0]


class ExampleTab(TabToolbox, example_tab_class):
    """Tab for plotting logging data"""

    _connected_signal = pyqtSignal(str)
    _disconnected_signal = pyqtSignal(str)
    _log_data_signal = pyqtSignal(int, object, object)
    _log_error_signal = pyqtSignal(object, str)
    _param_updated_signal = pyqtSignal(str, str)

    def __init__(self, helper):
        super(ExampleTab, self).__init__(helper, 'Example')
        self.setupUi(self)

        # Always wrap callbacks from Crazyflie API though QT Signal/Slots
        # to avoid manipulating the UI when rendering it
        self._connected_signal.connect(self._connected)
        self._disconnected_signal.connect(self._disconnected)
        self._log_data_signal.connect(self._log_data_received)
        self._param_updated_signal.connect(self._param_updated)

        # Connect the Crazyflie API callbacks to the signals
        self._helper.cf.connected.add_callback(
            self._connected_signal.emit)

        self._helper.cf.disconnected.add_callback(
            self._disconnected_signal.emit)

    def _connected(self, link_uri):
        """Callback when the Crazyflie has been connected"""

        logger.debug("Crazyflie connected to {}".format(link_uri))

    def _disconnected(self, link_uri):
        """Callback for when the Crazyflie has been disconnected"""

        logger.debug("Crazyflie disconnected from {}".format(link_uri))

    def _param_updated(self, name, value):
        """Callback when the registered parameter get's updated"""

        logger.debug("Updated {0} to {1}".format(name, value))

    def _log_data_received(self, timestamp, data, log_conf):
        """Callback when the log layer receives new data"""

        logger.debug("{0}:{1}:{2}".format(timestamp, log_conf.name, data))

    def _logging_error(self, log_conf, msg):
        """Callback from the log layer when an error occurs"""

        QMessageBox.about(self, "Example error",
                          "Error when using log config"
                          " [{0}]: {1}".format(log_conf.name, msg))



# -*- coding: utf-8 -*-
#
#     ||          ____  _ __
#  +------+      / __ )(_) /_______________ _____  ___
#  | 0xBC |     / __  / / __/ ___/ ___/ __ `/_  / / _ \
#  +------+    / /_/ / / /_/ /__/ /  / /_/ / / /_/  __/
#   ||  ||    /_____/_/\__/\___/_/   \__,_/ /___/\___/
#
#  Copyright (C) 2018-2023 Bitcraze AB
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either version 2
#  of the License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA  02110-1301, USA.

"""
Dialog box used to configure anchor positions. Used from the LPS tab.
"""
import logging

import cfclient
from cfclient.utils.logconfigreader import FILE_REGEX_YAML
from PyQt6 import QtWidgets
from PyQt6 import uic
from PyQt6.QtCore import QAbstractTableModel, QVariant, Qt
from PyQt6.QtGui import QBrush, QColor
from PyQt6.QtWidgets import QInputDialog, QFileDialog
import yaml
import os

__author__ = 'Bitcraze AB'
__all__ = ['AnchorPositionDialog']

logger = logging.getLogger(__name__)

(anchor_postiong_widget_class, connect_widget_base_class) = (
    uic.loadUiType(
        cfclient.module_path + '/ui/dialogs/anchor_position_dialog.ui')
)


class AnchorPositionConfigTableModel(QAbstractTableModel):
    def __init__(self, headers, parent=None, *args):
        QAbstractTableModel.__init__(self, parent)
        self._anchor_positions = []
        self._headers = headers
        self._latest_known_anchor_positions = {}

        self._green_brush = QBrush(QColor(200, 255, 200))
        self._red_brush = QBrush(QColor(255, 200, 200))

    def rowCount(self, parent=None, *args, **kwargs):
        return len(self._anchor_positions)

    def columnCount(self, parent=None, *args, **kwargs):
        return len(self._headers)

    def data(self, index, role=None):
        value = self._anchor_positions[index.row()][index.column()]
        if index.isValid():
            if index.column() == 0:
                if role == Qt.ItemDataRole.CheckStateRole:
                    return QVariant(value)
            elif index.column() == 1:
                if role == Qt.ItemDataRole.DisplayRole:
                    return QVariant(value)
            else:
                if role == Qt.ItemDataRole.DisplayRole:
                    return QVariant('%.2f' % (value))
                elif role == Qt.ItemDataRole.EditRole:
                    return QVariant(value)
                elif role == Qt.ItemDataRole.BackgroundRole:
                    return self._get_background(index.row(), index.column())

        return QVariant()

    def setData(self, index, value, role=Qt.ItemDataRole.EditRole):
        if not index.isValid():
            return False

        self._anchor_positions[index.row()][index.column()] = value
        return True

    def headerData(self, col, orientation, role=None):
        if orientation == Qt.Orientation.Horizontal and role == Qt.ItemDataRole.DisplayRole:
            return QVariant(self._headers[col])
        return QVariant()

    def flags(self, index):
        if not index.isValid():
            return None

        if index.column() == 0:
            return Qt.ItemFlag.ItemIsEnabled | Qt.ItemFlag.ItemIsUserCheckable
        elif index.column() == 1:
            return Qt.ItemFlag.ItemIsEnabled
        else:
            return Qt.ItemFlag.ItemIsEnabled | Qt.ItemFlag.ItemIsEditable

    def add_anchor(self, anchor_id, x=0.0, y=0.0, z=0.0):
        if not self._id_exist(anchor_id):
            self.layoutAboutToBeChanged.emit()
            self._anchor_positions.append([0, anchor_id, x, y, z])
            self._anchor_positions.sort(key=lambda row: row[1])
            self.layoutChanged.emit()

    def replace_anchors_from_latest_known_positions(self):
        self.replace_anchor_positions(self._latest_known_anchor_positions)

    def replace_anchor_positions(self, anchor_positions):
        self.layoutAboutToBeChanged.emit()
        self._anchor_positions = []
        for id, position in anchor_positions.items():
            self.add_anchor(id, x=position[0], y=position[1], z=position[2])
        self.layoutChanged.emit()

    def get_anchor_postions(self):
        result = {}
        for row in self._anchor_positions:
            result[row[1]] = (row[2], row[3], row[4])
        return result

    def anchor_postions_updated(self, anchor_positions):
        self.layoutAboutToBeChanged.emit()
        self._latest_known_anchor_positions = anchor_positions
        self.layoutChanged.emit()

    def remove_selected_anchors(self):
        self.layoutAboutToBeChanged.emit()
        self._anchor_positions = list(filter(
            lambda row: row[0] == 0, self._anchor_positions))
        self.layoutChanged.emit()

    def _id_exist(self, anchor_id):
        for anchor in self._anchor_positions:
            if anchor[1] == anchor_id:
                return True
        return False

    def _get_background(self, row, col):
        id = self._anchor_positions[row][1]
        if id in self._latest_known_anchor_positions:
            current_value = self._anchor_positions[row][col]
            latest_value = self._latest_known_anchor_positions[id][col - 2]

            if abs(current_value - latest_value) < 0.005:
                return self._green_brush
            else:
                return self._red_brush

        return QVariant()


class AnchorPositionDialog(QtWidgets.QWidget, anchor_postiong_widget_class):

    def __init__(self, lps_tab, helper):
        super(AnchorPositionDialog, self).__init__()
        self.setupUi(self)

        self._lps_tab = lps_tab
        self._helper = helper

        self._headers = ['', 'id', 'x', 'y', 'z']
        self._data_model = AnchorPositionConfigTableModel(self._headers, self)
        self._table_view.setModel(self._data_model)

        self._table_view.verticalHeader().setVisible(False)

        header = self._table_view.horizontalHeader()
        header.setSectionResizeMode(0, QtWidgets.QHeaderView.ResizeMode.ResizeToContents)
        header.setSectionResizeMode(1, QtWidgets.QHeaderView.ResizeMode.ResizeToContents)
        header.setSectionResizeMode(2, QtWidgets.QHeaderView.ResizeMode.Stretch)
        header.setSectionResizeMode(3, QtWidgets.QHeaderView.ResizeMode.Stretch)
        header.setSectionResizeMode(4, QtWidgets.QHeaderView.ResizeMode.Stretch)

        self._add_anchor_button.clicked.connect(
            self._add_anchor_button_clicked)
        self._remove_anchors_button.clicked.connect(
            self._data_model.remove_selected_anchors)
        self._get_from_anchors_button.clicked.connect(
            self._get_from_anchors_button_clicked)
        self._write_to_anchors_button.clicked.connect(
            self._write_to_anchors_button_clicked)
        self._close_button.clicked.connect(self.close)
        self._load_button.clicked.connect(
            self._load_button_clicked)
        self._save_button.clicked.connect(
            self._save_button_clicked)

    def _add_anchor_button_clicked(self):
        anchor_id, ok = QInputDialog.getInt(
            self, "New anchor", "Enter id", min=0, max=255)
        if ok:
            self._data_model.add_anchor(anchor_id)

    def _get_from_anchors_button_clicked(self):
        self._data_model.replace_anchors_from_latest_known_positions()

    def _write_to_anchors_button_clicked(self):
        anchor_positions = self._data_model.get_anchor_postions()
        self._lps_tab.write_positions_to_anchors(anchor_positions)

    def anchor_postions_updated(self, anchor_positions):
        self._data_model.anchor_postions_updated(anchor_positions)

    def _load_button_clicked(self):
        names = QFileDialog.getOpenFileName(self, 'Open file', self._helper.current_folder, FILE_REGEX_YAML)

        if names[0] == '':
            return

        self._helper.current_folder = os.path.dirname(names[0])

        f = open(names[0], 'r')
        with f:
            data = yaml.safe_load(f)

            anchor_positions = {}
            for id, pos in data.items():
                anchor_positions[id] = (pos['x'], pos['y'], pos['z'])
            self._data_model.replace_anchor_positions(anchor_positions)

    def _save_button_clicked(self):
        anchor_positions = self._data_model.get_anchor_postions()
        data = {}
        for id, pos in anchor_positions.items():
            data[id] = {'x': pos[0], 'y': pos[1], 'z': pos[2]}

        names = QFileDialog.getSaveFileName(self, 'Save file', self._helper.current_folder, FILE_REGEX_YAML)

        if names[0] == '':
            return

        self._helper.current_folder = os.path.dirname(names[0])

        if not names[0].endswith(".yaml") and names[0].find(".") < 0:
            filename = names[0] + ".yaml"
        else:
            filename = names[0]

        f = open(filename, 'w')
        with f:
            yaml.dump(data, f)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
#     ||          ____  _ __
#  +------+      / __ )(_) /_______________ _____  ___
#  | 0xBC |     / __  / / __/ ___/ ___/ __ `/_  / / _ \
#  +------+    / /_/ / / /_/ /__/ /  / /_/ / / /_/  __/
#   ||  ||    /_____/_/\__/\___/_/   \__,_/ /___/\___/
#
#  Copyright (C) 2011-2023 Bitcraze AB
#
#  Crazyflie Nano Quadcopter Client
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either version 2
#  of the License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.

#  You should have received a copy of the GNU General Public License along with
#  this program; if not, write to the Free Software Foundation, Inc.,
#  51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

"""
This class provides a spin box with hexadecimal numbers and arbitrarily length
(i.e. not limited by 32 bit).
"""

from PyQt6.QtGui import QRegularExpressionValidator
from PyQt6.QtCore import QRegularExpression
from PyQt6.QtCore import pyqtSignal
from PyQt6.QtWidgets import QAbstractSpinBox

__author__ = 'Bitcraze AB'
__all__ = ['HexSpinBox']


class HexSpinBox(QAbstractSpinBox):
    valueChanged = pyqtSignal(object)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        regexp = QRegularExpression('^0x[0-9A-Fa-f]{1,10}$')
        self.validator = QRegularExpressionValidator(regexp)
        self.setValue(0)

    def validate(self, text, pos):
        return self.validator.validate(text, pos)

    def textFromValue(self, value):
        return "0x%X" % value

    def valueFromText(self, text):
        return int(str(text), 0)

    def setValue(self, value):
        self._value = value
        self.lineEdit().setText(self.textFromValue(value))
        self.valueChanged.emit(self._value)

    def value(self):
        self._value = self.valueFromText(self.lineEdit().text())
        return self._value

    def stepBy(self, steps):
        self.setValue(self._value + steps)

    def stepEnabled(self):
        return (QAbstractSpinBox.StepEnabledFlag.StepUpEnabled | QAbstractSpinBox.StepEnabledFlag.StepDownEnabled)

    def is_text_different_from_value(self):
        return self._value != self.valueFromText(self.lineEdit().text())



# -*- coding: utf-8 -*-
#
#     ||          ____  _ __
#  +------+      / __ )(_) /_______________ _____  ___
#  | 0xBC |     / __  / / __/ ___/ ___/ __ `/_  / / _ \
#  +------+    / /_/ / / /_/ /__/ /  / /_/ / / /_/  __/
#   ||  ||    /_____/_/\__/\___/_/   \__,_/ /___/\___/
#
#  Copyright (C) 2021-2023 Bitcraze AB
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either version 2
#  of the License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA  02110-1301, USA.

"""
Dialog box used to configure base station geometry. Used from the lighthouse tab.
"""
import logging

import cfclient
from PyQt6 import QtWidgets
from PyQt6 import uic
from PyQt6.QtCore import QVariant, Qt, QAbstractTableModel, pyqtSignal
from cflib.localization import LighthouseSweepAngleAverageReader
from cflib.crazyflie.mem import LighthouseBsGeometry
from cfclient.ui.wizards.lighthouse_geo_bs_estimation_wizard import LighthouseBasestationGeometryWizard

__author__ = 'Bitcraze AB'
__all__ = ['LighthouseBsGeometryDialog']

logger = logging.getLogger(__name__)

(basestation_geometry_widget_class, connect_widget_base_class) = (
    uic.loadUiType(
        cfclient.module_path + '/ui/dialogs/lighthouse_bs_geometry_dialog.ui')
)


class LighthouseBsGeometryTableModel(QAbstractTableModel):
    def __init__(self, headers, parent=None, *args):
        QAbstractTableModel.__init__(self, parent)
        self._headers = headers
        self._table_values = []
        self._current_geos = {}
        self._estimated_geos = {}

    def rowCount(self, parent=None, *args, **kwargs):
        return len(self._table_values)

    def columnCount(self, parent=None, *args, **kwargs):
        return len(self._headers)

    def data(self, index, role=None):
        if index.isValid():
            value = self._table_values[index.row()][index.column()]
            if role == Qt.ItemDataRole.DisplayRole:
                return QVariant(value)

        return QVariant()

    def headerData(self, col, orientation, role=None):
        if orientation == Qt.Orientation.Horizontal and role == Qt.ItemDataRole.DisplayRole:
            return QVariant(self._headers[col])
        return QVariant()

    def _compile_entry(self, current_geo, estimated_geo, index):
        result = 'N/A'
        if current_geo is not None:
            result = '%.2f' % current_geo.origin[index]
        if estimated_geo is not None:
            result += ' -> %.2f' % estimated_geo.origin[index]

        return result

    def _add_table_value(self, current_geo, estimated_geo, id, table_values):
        x = self._compile_entry(current_geo, estimated_geo, 0)
        y = self._compile_entry(current_geo, estimated_geo, 1)
        z = self._compile_entry(current_geo, estimated_geo, 2)

        table_values.append([id + 1, x, y, z])

    def _add_table_value_for_id(self, current_geos, estimated_geos, table_values, id):
        current_geo = None
        if id in current_geos:
            current_geo = current_geos[id]

        estimated_geo = None
        if id in estimated_geos:
            estimated_geo = estimated_geos[id]

        if current_geo is not None or estimated_geo is not None:
            self._add_table_value(current_geo, estimated_geo, id, table_values)

    def _add_table_values(self, current_geos, estimated_geos, table_values):
        current_ids = set(current_geos.keys())
        estimated_ids = set(estimated_geos.keys())
        all_ids = current_ids.union(estimated_ids)

        for id in all_ids:
            self._add_table_value_for_id(current_geos, estimated_geos, table_values, id)

    def _update_table_data(self):
        self.layoutAboutToBeChanged.emit()
        self._table_values = []
        self._add_table_values(self._current_geos, self._estimated_geos, self._table_values)
        self._table_values.sort(key=lambda row: row[0])
        self.layoutChanged.emit()

    def set_estimated_geos(self, geos):
        self._estimated_geos = geos
        self._update_table_data()

    def set_current_geos(self, geos):
        self._current_geos = geos
        self._update_table_data()


class LighthouseBsGeometryDialog(QtWidgets.QWidget, basestation_geometry_widget_class):

    _sweep_angles_received_and_averaged_signal = pyqtSignal(object)
    _base_station_geometery_received_signal = pyqtSignal(object)

    def __init__(self, lighthouse_tab, *args):
        super(LighthouseBsGeometryDialog, self).__init__(*args)
        self.setupUi(self)

        self._lighthouse_tab = lighthouse_tab

        self._estimate_geometry_button.clicked.connect(self._estimate_geometry_button_clicked)

        self._write_to_cf_button.clicked.connect(self._write_to_cf_button_clicked)

        self._sweep_angles_received_and_averaged_signal.connect(self._sweep_angles_received_and_averaged_cb)
        self._base_station_geometery_received_signal.connect(self._basestation_geometry_received_signal_cb)
        self._close_button.clicked.connect(self.close)

        self._sweep_angle_reader = LighthouseSweepAngleAverageReader(
            self._lighthouse_tab._helper.cf, self._sweep_angles_received_and_averaged_signal.emit)

        self._base_station_geometry_wizard = LighthouseBasestationGeometryWizard(
            self._lighthouse_tab._helper.cf, self._base_station_geometery_received_signal.emit)

        self._lh_geos = None
        self._newly_estimated_geometry = {}

        # Table handlers
        self._headers = ['id', 'x', 'y', 'z']
        self._data_model = LighthouseBsGeometryTableModel(self._headers, self)
        self._table_view.setModel(self._data_model)

        self._table_view.verticalHeader().setVisible(False)

        header = self._table_view.horizontalHeader()
        header.setSectionResizeMode(0, QtWidgets.QHeaderView.ResizeMode.ResizeToContents)
        header.setSectionResizeMode(1, QtWidgets.QHeaderView.ResizeMode.Stretch)
        header.setSectionResizeMode(2, QtWidgets.QHeaderView.ResizeMode.Stretch)
        header.setSectionResizeMode(3, QtWidgets.QHeaderView.ResizeMode.Stretch)

        self._update_ui()

    def reset(self):
        self._newly_estimated_geometry = {}
        self._update_ui()

    def _basestation_geometry_received_signal_cb(self, basestation_geometries):
        self._newly_estimated_geometry = basestation_geometries
        self.show()
        self._update_ui()

    def _sweep_angles_received_and_averaged_cb(self, averaged_angles):
        self._averaged_angles = averaged_angles
        self._newly_estimated_geometry = {}

        for id, average_data in averaged_angles.items():
            sensor_data = average_data[1]
            rotation_bs_matrix, position_bs_vector = self._simple_estimator.estimate_geometry(sensor_data)
            geo = LighthouseBsGeometry()
            geo.rotation_matrix = rotation_bs_matrix
            geo.origin = position_bs_vector
            geo.valid = True
            self._newly_estimated_geometry[id] = geo

        self._update_ui()

    def _estimate_geometry_button_clicked(self):
        self._base_station_geometry_wizard.reset()
        self._base_station_geometry_wizard.show()
        self.hide()

    def _write_to_cf_button_clicked(self):
        if len(self._newly_estimated_geometry) > 0:
            self._lighthouse_tab.write_and_store_geometry(self._newly_estimated_geometry)
            self._newly_estimated_geometry = {}

        self._update_ui()

    def _update_ui(self):
        self._write_to_cf_button.setEnabled(len(self._newly_estimated_geometry) > 0)
        self._data_model.set_estimated_geos(self._newly_estimated_geometry)

    def closeEvent(self, event):
        self._stop_collection()

    def _stop_collection(self):
        self._sweep_angle_reader.stop_angle_collection()

    def geometry_updated(self, geometry):
        self._data_model.set_current_geos(geometry)

        self._update_ui()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
#     ||          ____  _ __
#  +------+      / __ )(_) /_______________ _____  ___
#  | 0xBC |     / __  / / __/ ___/ ___/ __ `/_  / / _ \
#  +------+    / /_/ / / /_/ /__/ /  / /_/ / / /_/  __/
#   ||  ||    /_____/_/\__/\___/_/   \__,_/ /___/\___/
#
#  Copyright (C) 2011-2023 Bitcraze AB
#
#  Crazyflie Nano Quadcopter Client
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either version 2
#  of the License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.

#  You should have received a copy of the GNU General Public License along with
#  this program; if not, write to the Free Software Foundation, Inc.,
#  51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

"""
The bootloader dialog is used to update the Crazyflie firmware and to
read/write the configuration block in the Crazyflie flash.
"""
from __future__ import annotations

from cflib.bootloader import Bootloader
from cfclient.ui.connectivity_manager import ConnectivityManager

import tempfile
import logging
import json
import os
import re
import threading
from urllib.request import urlopen
from urllib.error import URLError
import zipfile

from PyQt6 import QtWidgets, uic
from PyQt6.QtCore import pyqtSlot, pyqtSignal, QThread, Qt
from PyQt6.QtGui import QPixmap

import cfclient
import cflib.crazyflie

__author__ = 'Bitcraze AB'
__all__ = ['BootloaderDialog']

logger = logging.getLogger(__name__)

service_dialog_class = uic.loadUiType(cfclient.module_path +
                                      "/ui/dialogs/bootloader.ui")[0]

# This url is used to fetch all the releases from the FirmwareDownloader
RELEASE_URL = 'https://api.github.com/repos/bitcraze/'\
              'crazyflie-release/releases'


class BootloaderDialog(QtWidgets.QWidget, service_dialog_class):
    """Tab for update the Crazyflie firmware and for reading/writing the config
    block in flash"""

    class UIState:
        DISCONNECTED = 0
        COLD_CONNECTING = 1
        COLD_CONNECTED = 2
        FW_CONNECTING = 3
        FW_CONNECTED = 4
        FW_SCANNING = 5
        FLASHING = 6
        RESET = 7

    _release_firmwares_found = pyqtSignal(object)
    _release_downloaded = pyqtSignal(str, object)

    def __init__(self, helper, *args):
        super(BootloaderDialog, self).__init__(*args)
        self.setupUi(self)

        self.tabName = "Service"

        # self.tabWidget = tabWidget
        self._helper = helper

        # self.cf = crazyflie
        self.clt = CrazyloadThread(self._helper.cf)

        # Connecting GUI signals (a pity to do that manually...)
        self.imagePathBrowseButton.clicked.connect(self.pathBrowse)
        self.programButton.clicked.connect(self.programAction)
        self.coldBootButton.clicked.connect(self.initiateColdboot)
        self.resetButton.clicked.connect(self.resetCopter)

        self._helper.connectivity_manager.register_ui_elements(
            ConnectivityManager.UiElementsContainer(
                interface_combo=self.comboBox,
                address_spinner=self.address,
                connect_button=self.connectButton,
                scan_button=self.scanButton))
        self._helper.connectivity_manager.connection_state_changed.connect(self._fw_connection_state_changed)

        # connecting other signals
        self.clt.programmed.connect(self.programDone)
        self.clt.statusChanged.connect(self.statusUpdate)
        # self.clt.updateBootloaderStatusSignal.connect(
        #                                         self.updateBootloaderStatus)
        self.clt.connectingSignal.connect(lambda: self.setUiState(self.UIState.COLD_CONNECTING))

        self.clt.connectedSignal.connect(lambda: self._load_thread_connection_event(self.UIState.COLD_CONNECTED))
        self.clt.disconnectedSignal.connect(lambda: self._load_thread_connection_event(self.UIState.DISCONNECTED))
        self.clt.failed_signal.connect(lambda m: self._ui_connection_fail(m))

        self._cold_boot_error_message = None
        self._state = self.UIState.DISCONNECTED

        self._releases = {}
        self._platform_widget_names = {}
        self._release_firmwares_found.connect(self._populate_firmware_dropdown)
        self._release_downloaded.connect(self.release_zip_downloaded)
        self.firmware_downloader = FirmwareDownloader(self._release_firmwares_found, self._release_downloaded)
        self.firmware_downloader.get_firmware_releases()

        self.firmware_downloader.start()
        self.clt.start()

        self._platform_filter_checkboxes = []

        self._set_image(self.image_1, "src/cfclient/ui/icons/bolt.webp")
        self._set_image(self.image_2, "src/cfclient/ui/icons/cf21.webp")
        self._set_image(self.image_3, "src/cfclient/ui/icons/bl.webp")
        self._set_image(self.image_4, "src/cfclient/ui/icons/flapper.webp")
        self._set_image(self.image_5, "src/cfclient/ui/icons/tag.webp")

    def _ui_connection_fail(self, message):
        self._cold_boot_error_message = message
        self.setUiState(self.UIState.DISCONNECTED)

    def _set_image(self, image_label, image_path):
        IMAGE_WIDTH = 100
        IMAGE_HEIGHT = 100

        pixmap = QPixmap(image_path)

        if pixmap.isNull():
            logger.warning(f"Failed to load image: {image_path}")
            image_label.setText("Missing image")
        else:
            scaled_pixmap = pixmap.scaled(
                IMAGE_WIDTH, IMAGE_HEIGHT,
                Qt.AspectRatioMode.KeepAspectRatio,
                Qt.TransformationMode.SmoothTransformation
            )
            image_label.setPixmap(scaled_pixmap)

    def setUiState(self, state):
        self._state = state

        if (state == self.UIState.DISCONNECTED):
            self.resetButton.setEnabled(False)
            self.programButton.setEnabled(False)
            if self._cold_boot_error_message is not None:
                self.setStatusLabel(self._cold_boot_error_message)
            else:
                self.setStatusLabel("Not connected")
            self.coldBootButton.setEnabled(True)
            self.progressBar.setTextVisible(False)
            self.progressBar.setValue(0)
            self.statusLabel.setText('Status: <b>IDLE</b>')
            self.setSourceSelectionUiEnabled(True)
            self._helper.connectivity_manager.set_enable(True)
        elif (state == self.UIState.COLD_CONNECTING):
            self._cold_boot_error_message = None
            self.resetButton.setEnabled(False)
            self.programButton.setEnabled(False)
            self.setStatusLabel("Trying to connect cold bootloader, restart the Crazyflie to connect")
            self.coldBootButton.setEnabled(False)
            self.setSourceSelectionUiEnabled(True)
            self._helper.connectivity_manager.set_enable(False)
        elif (state == self.UIState.COLD_CONNECTED):
            self._cold_boot_error_message = None
            self.resetButton.setEnabled(True)
            if any(button.isChecked() for button in self._platform_filter_checkboxes):
                self.programButton.setEnabled(True)
            else:
                self.programButton.setToolTip("Select a platform before programming.")
            self.setStatusLabel("Connected to bootloader")
            self.coldBootButton.setEnabled(False)
            self.imagePathBrowseButton.setEnabled(True)
            self.imagePathLine.setEnabled(True)
            self.firmwareDropdown.setEnabled(True)
            self._helper.connectivity_manager.set_enable(False)
        elif (state == self.UIState.FW_CONNECTING):
            self._cold_boot_error_message = None
            self.resetButton.setEnabled(False)
            self.programButton.setEnabled(False)
            self.setStatusLabel("Trying to connect in firmware mode")
            self.coldBootButton.setEnabled(False)
            self.setSourceSelectionUiEnabled(True)
            self._helper.connectivity_manager.set_enable(True)
        elif (state == self.UIState.FW_CONNECTED):
            self._cold_boot_error_message = None
            self.resetButton.setEnabled(False)
            self.coldBootButton.setEnabled(False)
            self._helper.connectivity_manager.set_enable(True)

            if self._helper.cf.link_uri.startswith("usb://"):
                self.programButton.setEnabled(False)
                self.setStatusLabel("Connected using USB")
                self.setSourceSelectionUiEnabled(False)
            else:
                if any(button.isChecked() for button in self._platform_filter_checkboxes):
                    self.programButton.setEnabled(True)
                else:
                    self.programButton.setToolTip("Select a platform before programming.")
                self.setStatusLabel("Connected in firmware mode")
                self.setSourceSelectionUiEnabled(True)
        elif (state == self.UIState.FW_SCANNING):
            self._cold_boot_error_message = None
            self.resetButton.setEnabled(False)
            self.programButton.setEnabled(False)
            self.setStatusLabel("Scanning")
            self.coldBootButton.setEnabled(False)
            self.setSourceSelectionUiEnabled(True)
            self._helper.connectivity_manager.set_enable(True)
        elif (state == self.UIState.FLASHING):
            self.resetButton.setEnabled(False)
            self.resetButton.setEnabled(False)
            self.programButton.setEnabled(False)
            self.setStatusLabel("Flashing")
            self.coldBootButton.setEnabled(False)
            self.setSourceSelectionUiEnabled(False)
            self._helper.connectivity_manager.set_enable(False)
        elif (state == self.UIState.RESET):
            self._cold_boot_error_message = None
            self.setStatusLabel("Resetting to firmware, disconnected")
            self.resetButton.setEnabled(False)
            self.programButton.setEnabled(False)
            self.coldBootButton.setEnabled(False)
            self.setSourceSelectionUiEnabled(True)
            self._helper.connectivity_manager.set_enable(False)
        self._update_program_button_state()

    def setSourceSelectionUiEnabled(self, enabled):
        self.imagePathBrowseButton.setEnabled(enabled)
        self.imagePathLine.setEnabled(enabled)
        self.firmwareDropdown.setEnabled(enabled)

    def setStatusLabel(self, text):
        self.connectionStatus.setText("Status: <b>%s</b>" % text)

    def resetCopter(self):
        self.clt.resetCopterSignal.emit()
        self.setUiState(self.UIState.RESET)

    def updateConfig(self, channel, speed, rollTrim, pitchTrim):
        self.rollTrim.setValue(rollTrim)
        self.pitchTrim.setValue(pitchTrim)
        self.radioChannel.setValue(channel)
        self.radioSpeed.setCurrentIndex(speed)

    def closeEvent(self, event):
        self.clt.terminate_flashing()
        # Remove downloaded-firmware files.
        self.firmware_downloader.bootload_complete.emit()

    def _populate_firmware_dropdown(self, releases):
        """ Callback from firmware-downloader that retrieves all
            the latest firmware-releases.
        """
        platforms = set()
        for release in releases:
            release_name = release[0]
            downloads = release[1:]

            downloads.sort(key=self.download_sorter)

            for download in downloads:
                download_name, download_link = download
                platform = self._extract_platform(download_name)
                # Ignore old releases that do not use the standard file naming convention
                if platform:
                    widget_name = '%s - %s' % (release_name, download_name)
                    if platform not in self._platform_widget_names:
                        self._platform_widget_names[platform] = []
                    self._platform_widget_names[platform].append(widget_name)
                    self._releases[widget_name] = download_link

                    platforms.add(platform)

        for platform in sorted(platforms, reverse=True):
            RADIO_BUTTON_WIDTH = 100

            radio_button = QtWidgets.QRadioButton(platform)

            radio_button.setFixedWidth(RADIO_BUTTON_WIDTH)

            radio_button.toggled.connect(self._update_firmware_dropdown)
            radio_button.toggled.connect(self._update_program_button_state)

            self._platform_filter_checkboxes.append(radio_button)
            self.filterLayout.insertWidget(0, radio_button)

        self.firmwareDropdown.setSizeAdjustPolicy(QtWidgets.QComboBox.SizeAdjustPolicy.AdjustToContents)
        self._update_firmware_dropdown(True)

    def _update_program_button_state(self):
        any_checked = any(button.isChecked() for button in self._platform_filter_checkboxes)
        is_connected = self._state in [
            self.UIState.COLD_CONNECTED,
            self.UIState.FW_CONNECTED
        ]

        self.programButton.setEnabled(any_checked and is_connected)

        if not is_connected:
            self.programButton.setToolTip("Connect your device before programming.")
        elif not any_checked:
            self.programButton.setToolTip("Select a platform before programming.")
        else:
            self.programButton.setToolTip("")

    def _update_firmware_dropdown(self, active: bool):
        if active:
            platform = None
            for button in self._platform_filter_checkboxes:
                if button.isChecked():
                    platform = button.text()

            if platform:
                self.firmwareDropdown.clear()
                for widget_name in self._platform_widget_names[platform]:
                    self.firmwareDropdown.addItem(widget_name)

    def _extract_platform(self, download_name: str) -> str | None:
        # Download name is something like 'firmware-cf2-2022.12.zip'
        found = re.search('firmware-(\\w+)-', download_name)
        if found:
            groups = found.groups()
            if len(groups) == 1:
                return groups[0]
        return None

    def download_sorter(self, element):
        '''Sort downloads to display cf2 before bolt and tag'''
        name = element[0]
        if 'cf2' in name:
            return '0' + name
        else:
            return '1' + name

    def release_zip_downloaded(self, release_name, release_path):
        """ Callback when a release is successfully downloaded and
            save to release_path.
        """
        self.downloadStatus.setText('Downloaded')
        self.clt.program.emit(release_path, '')

    def _load_thread_connection_event(self, new_sate):
        if self._state != self.UIState.FLASHING:
            self.setUiState(new_sate)

    def _fw_connection_state_changed(self, new_state):
        if self._state != self.UIState.FLASHING:
            if new_state == ConnectivityManager.UIState.DISCONNECTED:
                self.setUiState(self.UIState.DISCONNECTED)
            elif new_state == ConnectivityManager.UIState.CONNECTING:
                self.setUiState(self.UIState.FW_CONNECTING)
            elif new_state == ConnectivityManager.UIState.CONNECTED:
                self.setUiState(self.UIState.FW_CONNECTED)
            elif new_state == ConnectivityManager.UIState.SCANNING:
                self.setUiState(self.UIState.FW_SCANNING)

    @pyqtSlot()
    def pathBrowse(self):
        names = QtWidgets.QFileDialog.getOpenFileName(
            self, 'Release file to flash', self._helper.current_folder, "*.zip")
        if names[0] == '':
            return

        filename = names[0]
        self._helper.current_folder = os.path.dirname(filename)

        if filename.endswith('.zip'):
            self.imagePathLine.setText(filename)
        else:
            msgBox = QtWidgets.QMessageBox()
            msgBox.setText("Wrong file extention. Must be .zip.")
            msgBox.exec_()

    @pyqtSlot()
    def programAction(self):
        if self._state == self.UIState.COLD_CONNECTED:
            self.clt.set_boot_mode(self.clt.COLD_BOOT)
        else:
            self.clt.set_boot_mode(self.clt.WARM_BOOT)

        # call the flasher
        if self.sourceTab.currentWidget() == self.tabFromFile:
            if self.imagePathLine.text() == "":
                msgBox = QtWidgets.QMessageBox()
                msgBox.setText("Please choose an image file to program.")
                msgBox.exec_()

                return

            self.setUiState(self.UIState.FLASHING)

            # by default, flash everything in the zip (if possible)
            mcu_to_flash = None
            self.clt.program.emit(self.imagePathLine.text(), mcu_to_flash)
        else:
            self.setUiState(self.UIState.FLASHING)

            requested_release = self.firmwareDropdown.currentText()
            download_url = self._releases[requested_release]
            self.downloadStatus.setText('Fetching...')
            self.firmware_downloader.download_release(requested_release, download_url)

    @pyqtSlot(bool)
    def programDone(self, success):
        if success:
            self.statusLabel.setText('Status: <b>Programing complete!</b>')
            self.downloadStatus.setText('')
        else:
            self.statusLabel.setText('Status: <b>Programing failed!</b>')

        self.setUiState(self.UIState.DISCONNECTED)
        self.resetCopter()

    @pyqtSlot(str, int)
    def statusUpdate(self, status, progress):
        logger.debug("Status: [%s] | %d", status, progress)
        self.statusLabel.setText('Status: <b>' + status + '</b>')
        if progress >= 0:
            self.progressBar.setValue(int(progress))

    def initiateColdboot(self):
        self.clt.initiateColdBootSignal.emit("radio://0/100")


# No run method specified here as the default run implementation is running the
# event loop which is what we want
class CrazyloadThread(QThread):
    # Input signals declaration (not sure it should be used like that...)
    program = pyqtSignal(str, str)
    initiateColdBootSignal = pyqtSignal(str)
    resetCopterSignal = pyqtSignal()
    writeConfigSignal = pyqtSignal(int, int, float, float)
    # Output signals declaration
    programmed = pyqtSignal(bool)
    verified = pyqtSignal()
    statusChanged = pyqtSignal(str, int)
    connectedSignal = pyqtSignal()
    connectingSignal = pyqtSignal()
    failed_signal = pyqtSignal(str)
    disconnectedSignal = pyqtSignal()
    updateConfigSignal = pyqtSignal(int, int, float, float)
    updateCpuIdSignal = pyqtSignal(str)

    radioSpeedPos = 2

    WARM_BOOT = 0
    COLD_BOOT = 1

    def __init__(self, crazyflie: cflib.crazyflie.Crazyflie):
        super(CrazyloadThread, self).__init__()

        self._terminate_flashing = False

        self._bl = Bootloader()

        self._cf = crazyflie

        self._boot_mode = self.COLD_BOOT

        # Make sure that the signals are handled by this thread event loop
        self.moveToThread(self)

        self.program.connect(self.programAction)
        self.initiateColdBootSignal.connect(self.initiateColdBoot)
        self.resetCopterSignal.connect(self.resetCopter)

    def __del__(self):
        self.quit()
        self.wait()

    def set_boot_mode(self, mode):
        self._boot_mode = mode

    def initiateColdBoot(self, linkURI):
        self.connectingSignal.emit()

        try:
            success = self._bl.start_bootloader(warm_boot=False, cf=None)
            if not success:
                self.failed_signal.emit("Could not connect to bootloader")
            else:
                self.connectedSignal.emit()
        except Exception as e:
            self.failed_signal.emit("{}".format(e))

    def programAction(self, filename, mcu_to_flash):
        targets = {}
        if mcu_to_flash:
            targets[mcu_to_flash] = ("fw",)
        try:
            self._terminate_flashing = False
            self._bl.clink = self._cf.link_uri
            self._bl.flash_full(self._cf,
                                str(filename),
                                self._boot_mode is self.WARM_BOOT,
                                targets,
                                None,
                                self.statusChanged.emit,
                                lambda: self._terminate_flashing)
            self.programmed.emit(True)
        except Exception as e:
            self.failed_signal.emit("{}".format(e))
            self.programmed.emit(False)

    def terminate_flashing(self):
        self._terminate_flashing = True

    def resetCopter(self):
        try:
            self._bl.reset_to_firmware()
        except Exception:
            pass
        self._bl.close()
        self.disconnectedSignal.emit()


class FirmwareDownloader(QThread):
    """ Uses github API to retrieves firmware-releases. """

    bootload_complete = pyqtSignal()

    def __init__(self, qtsignal_get_all_firmwares, qtsignal_get_release):
        super(FirmwareDownloader, self).__init__()

        self._qtsignal_get_all_firmwares = qtsignal_get_all_firmwares
        self._qtsignal_get_release = qtsignal_get_release

        self._tempDirectory = tempfile.TemporaryDirectory()

        self.moveToThread(self)

    def get_firmware_releases(self):
        """ Wrapper-function """
        threading.Thread(target=self._get_firmware_releases,
                         args=(self._qtsignal_get_all_firmwares, )).start()

    def download_release(self, release_name, url):
        """ Wrapper-function """
        threading.Thread(target=self._download_release,
                         args=(self._qtsignal_get_release,
                               release_name, url)).start()

    def _get_firmware_releases(self, signal):
        """ Gets the firmware releases from the github API
            and returns a list of format [rel-name, {release: download-link}].
            Returns None if the request fails.
        """
        response = {}
        try:
            with urlopen(RELEASE_URL) as resp:
                response = json.load(resp)
        except URLError:
            logger.warning(
                'Failed to make web request to get firmware-release')

        release_list = []

        for release in response:
            release_name = release['name']
            if release_name:
                releases = [release_name]
                for download in release['assets']:
                    releases.append((download['name'], download['browser_download_url']))
                release_list.append(releases)

        if release_list:
            signal.emit(release_list)
        else:
            logger.warning('Failed to parse firmware-releases in web request')

    def _download_release(self, signal, release_name, url):
        """ Downloads the given release and calls the callback signal
            if successful.
        """
        filepath = os.path.join(self._tempDirectory.name, release_name.split(' ')[-1])
        try:
            # Check if we have an old file saved and if so, ensure it's a valid
            # zipfile and then call signal
            with open(filepath, 'rb') as f:
                previous_release = zipfile.ZipFile(f)
                # testzip returns None if it's OK.
                if previous_release.testzip() is None:
                    logger.info('Using same firmware-release file at'
                                '%s' % filepath)
                    signal.emit(release_name, filepath)
                    return
        except FileNotFoundError:
            try:
                # Fetch the file with a new web request and save it to
                # a temporary file.
                with urlopen(url) as response:
                    with open(filepath, 'wb') as release_file:
                        release_file.write(response.read())
                    logger.info('Created temporary firmware-release file at'
                                '%s' % filepath)
                    signal.emit(release_name, filepath)
            except URLError:
                logger.warning('Failed to make web request to get requested'
                               ' firmware-release')



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
#     ||          ____  _ __
#  +------+      / __ )(_) /_______________ _____  ___
#  | 0xBC |     / __  / / __/ ___/ ___/ __ `/_  / / _ \
#  +------+    / /_/ / / /_/ /__/ /  / /_/ / / /_/  __/
#   ||  ||    /_____/_/\__/\___/_/   \__,_/ /___/\___/
#
#  Copyright (C) 2011-2025 Bitcraze AB
#
#  Crazyflie Nano Quadcopter Client
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either version 2
#  of the License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.

#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
#  02110-1301, USA.

"""
Basic tab to be able to set (and test) color in Color LED.
"""

import logging
from PyQt6 import uic
from PyQt6.QtCore import Qt, pyqtSignal
from PyQt6.QtGui import QColor, QPixmap, QPainter, QLinearGradient, QPen, QPainterPath
from PyQt6.QtWidgets import QPushButton, QMessageBox
from cflib.crazyflie.log import LogConfig

import cfclient
from cfclient.ui.tab_toolbox import TabToolbox
from cfclient.utils.config import Config

__author__ = 'Bitcraze AB'
__all__ = ['ColorLEDTab']

logger = logging.getLogger(__name__)

color_led_tab_class = uic.loadUiType(cfclient.module_path + "/ui/tabs/colorLEDTab.ui")[0]  # type: ignore


class wrgb_t:
    def __init__(self, w: int, r: int, g: int, b: int):
        self.w = w
        self.r = r
        self.g = g
        self.b = b

    def pack(self) -> int:
        """Pack WRGB values into uint32 format: 0xWWRRGGBB"""
        return (self.w << 24) | (self.r << 16) | (self.g << 8) | self.b


class rgb_t:
    def __init__(self, r: int, g: int, b: int):
        self.r = r
        self.g = g
        self.b = b

    def __iter__(self):
        return iter([self.r, self.g, self.b])

    def extract_white(self) -> wrgb_t:
        """Extract white channel and return WRGB color"""
        white = min(self.r, self.g, self.b)
        return wrgb_t(white, self.r - white, self.g - white, self.b - white)


class ThermalMonitor:
    """Monitors thermal throttling status for Color LED decks"""

    def __init__(self, helper, params_config, log_data_signal, log_error_signal):
        """
        Initialize thermal monitor

        Args:
            helper: Crazyflie helper instance
            params_config: Dictionary mapping positions to their thermal log names
            log_data_signal: PyQt signal to emit when log data is received
            log_error_signal: PyQt signal to emit when log errors occur
        """
        self._helper = helper
        self._params_config = params_config
        self._log_data_signal = log_data_signal
        self._log_error_signal = log_error_signal
        self._active_positions = []

    def start_monitoring(self, positions):
        """
        Start thermal monitoring for specified deck positions

        Args:
            positions: List of position indices to monitor
        """
        self._active_positions = positions

        for position in positions:
            if position not in self._params_config:
                continue

            params = self._params_config[position]
            log_name = f"Thermal{['Bottom', 'Top'][position]}"
            lg = LogConfig(log_name, Config().get("ui_update_period"))

            try:
                lg.add_variable(params['thermal_log'], "uint8_t")
                self._helper.cf.log.add_config(lg)
                lg.data_received_cb.add_callback(self._log_data_signal.emit)
                lg.error_cb.add_callback(self._log_error_signal.emit)
                lg.start()
                logger.debug(f"Started thermal logging for position {position}: {params['thermal_log']}")
            except (KeyError, AttributeError) as e:
                logger.debug(f"Could not start thermal logging for position {position}: {e}")

    def check_throttling(self, data, positions_to_check):
        """
        Check if any of the specified positions are thermally throttling

        Args:
            data: Log data dictionary
            positions_to_check: List of position indices to check

        Returns:
            bool: True if any position is throttling
        """
        return any(
            self._params_config[pos]['thermal_log'] in data and
            data[self._params_config[pos]['thermal_log']]
            for pos in positions_to_check
            if pos in self._params_config
        )


class ColorLEDDeckController:
    """Manages Color LED deck detection and parameter writes"""

    def __init__(self, helper, params_config):
        """
        Initialize deck controller

        Args:
            helper: Crazyflie helper instance
            params_config: Dictionary mapping positions to their parameter names
        """
        self._helper = helper
        self._params_config = params_config
        self._deck_present = {}  # Maps position -> bool

    def detect_decks(self):
        """Detect which Color LED decks are present"""
        for position, params in self._params_config.items():
            try:
                deck_param = self._helper.cf.param.get_value(params['deck_param'])
                # Convert to int to handle string values like "0" or "1"
                self._deck_present[position] = bool(int(deck_param))
                logger.debug(f"Color LED deck at position {position} ({'Bottom' if position == 0 else 'Top'}) "
                             f"detected: {self._deck_present[position]} (param: {params['deck_param']}={deck_param})")
            except (KeyError, ValueError, TypeError) as e:
                self._deck_present[position] = False
                logger.debug(f"Color LED deck parameter not found for position {position}: {e}")

    def is_deck_present(self, position):
        """Check if deck at given position is present"""
        return self._deck_present.get(position, False)

    def get_present_decks(self):
        """Get list of positions where decks are present"""
        return [pos for pos, present in self._deck_present.items() if present]

    def write_color(self, position, color_uint32):
        """
        Write color to a specific deck position

        Args:
            position: Deck position (0=bottom, 1=top)
            color_uint32: Packed WRGB color value
        """
        if position not in self._params_config:
            logger.warning(f"Unknown position {position}")
            return

        if not self.is_deck_present(position):
            logger.debug(f"Color LED deck at position {position} not present, skipping color write")
            return

        param_name = self._params_config[position]['color']
        self._helper.cf.param.set_value(param_name, str(color_uint32))

    def clear_deck_state(self):
        """Clear deck presence state (called on disconnect)"""
        self._deck_present.clear()


class ColorLEDTab(TabToolbox, color_led_tab_class):
    """Tab with inline color picker with hue slider, SV area, and hex input."""

    _colorChanged = pyqtSignal(QColor)
    _connectedSignal = pyqtSignal(str)
    _disconnectedSignal = pyqtSignal(str)
    _log_data_signal = pyqtSignal(int, object, object)
    _log_error_signal = pyqtSignal(object, str)

    # Parameter and log names by position
    PARAMS_BY_POSITION = {
        0: {  # Bottom
            'color': 'colorLedBot.wrgb8888',
            'thermal_log': 'colorLedBot.throttlePct',
            'brightness': 'colorLedBot.brightCorr',
            'deck_param': 'deck.bcColorLedBot'
        },
        1: {  # Top
            'color': 'colorLedTop.wrgb8888',
            'thermal_log': 'colorLedTop.throttlePct',
            'brightness': 'colorLedTop.brightCorr',
            'deck_param': 'deck.bcColorLedTop'
        }
    }

    def __init__(self, helper):
        super(ColorLEDTab, self).__init__(helper, 'Color LED')
        self.setupUi(self)

        self.groupBox_color.setEnabled(False)
        self.hue_bar.setEnabled(False)

        self._populate_position_dropdown()

        self._hue = 0
        self._saturation = 0
        self._value = 0

        self.hue_bar.setMinimum(0)
        self.hue_bar.setMaximum(1000)
        self.hue_bar.setValue(int(self._hue * 1000))
        self.hue_bar.valueChanged.connect(self._on_hue_slider_changed)
        self.hue_bar.setMouseTracking(True)

        self.hex_input.editingFinished.connect(self._on_hex_changed)

        self.custom_color_buttons = []
        self._connect_color_buttons()

        self.add_color_button.clicked.connect(self._add_custom_color_button)

        self.sv_area.setMouseTracking(True)
        self.sv_area.setStyleSheet("""
            QLabel {
                border-radius: 8px;
                border: 1px solid #444;
                background-color: #222;
            }
        """)

        self._isConnected = False
        self._updating_from_fetch = False  # Flag to prevent writes during color fetch
        self._throttling_state = {}  # Track last known throttling state per position

        # Initialize thermal monitor and deck controller
        self._thermal_monitor = ThermalMonitor(
            self._helper,
            self.PARAMS_BY_POSITION,
            self._log_data_signal,
            self._log_error_signal
        )
        self._deck_controller = ColorLEDDeckController(
            self._helper,
            self.PARAMS_BY_POSITION
        )

        self._connectedSignal.connect(self._connected)
        self._disconnectedSignal.connect(self._disconnected)
        self._colorChanged.connect(self._write_color_parameter)
        self._log_data_signal.connect(self._log_data_received)
        self._log_error_signal.connect(self._logging_error)

        self._helper.cf.connected.add_callback(self._connectedSignal.emit)
        self._helper.cf.disconnected.add_callback(self._disconnectedSignal.emit)

        # Handle position dropdown changes
        self.positionDropdown.currentIndexChanged.connect(self._on_position_changed)

    def _logging_error(self, log_conf, msg):
        QMessageBox.about(self, "Log error",
                          "Error when starting log config [%s]: %s" % (
                              log_conf.name, msg))

    def _log_data_received(self, _timestamp, data, _logconf):
        if not self.isVisible():
            return

        position = self.positionDropdown.currentData()

        # Determine which positions to check for throttling
        positions_to_check = [0, 1] if position == 2 else [position]

        # Update throttling state for positions that have data in this update
        has_thermal_data = False
        for pos in positions_to_check:
            if pos in self.PARAMS_BY_POSITION:
                thermal_log = self.PARAMS_BY_POSITION[pos]['thermal_log']
                if thermal_log in data:
                    has_thermal_data = True
                    # Update state for this position
                    self._throttling_state[pos] = bool(data[thermal_log])

        if has_thermal_data:
            # Check if any selected position is currently throttling based on known state
            is_throttling = any(
                self._throttling_state.get(pos, False)
                for pos in positions_to_check
            )

            self.information_text.setText(
                "Throttling: Lowering intensity to lower temperature." if is_throttling else ""
            )

    def _connected(self, _):
        self._isConnected = True

        # Detect which color LED decks are attached
        self._deck_controller.detect_decks()

        # Update dropdown based on detected decks
        self._update_position_dropdown()

        # Enable UI controls only if at least one deck is present
        present_decks = self._deck_controller.get_present_decks()
        has_decks = len(present_decks) > 0
        self.groupBox_color.setEnabled(has_decks)
        self.hue_bar.setEnabled(has_decks)

        # Set up thermal logging for available decks
        if has_decks:
            self._thermal_monitor.start_monitoring(present_decks)

    def _disconnected(self, _):
        self._isConnected = False
        self._deck_controller.clear_deck_state()
        self._throttling_state.clear()  # Clear throttling state

        # Disable UI controls
        self.groupBox_color.setEnabled(False)
        self.hue_bar.setEnabled(False)

        # Disable all position dropdown items
        self.positionDropdown.model().item(0).setEnabled(False)
        self.positionDropdown.model().item(1).setEnabled(False)
        self.positionDropdown.model().item(2).setEnabled(False)

        self.information_text.setText("")  # clear thermal throttling warning

    def _write_color_parameter(self, color: QColor):
        if not self._isConnected:
            return

        # Don't write when we're updating UI from a fetch operation
        if self._updating_from_fetch:
            return

        r, g, b, _ = color.getRgb()
        rgb = rgb_t(r or 0, g or 0, b or 0)
        wrgb = rgb.extract_white()
        color_uint32 = wrgb.pack()

        position = self.positionDropdown.currentData()

        # Determine which positions to write to
        positions_to_write = [0, 1] if position == 2 else [position]

        for pos in positions_to_write:
            self._deck_controller.write_color(pos, color_uint32)

    def _on_position_changed(self, _):
        """Handle position dropdown changes by fetching current color from Crazyflie"""
        if not self._isConnected:
            return

        position = self.positionDropdown.currentData()

        if position == 2:  # Both
            # Fetch both colors and compare
            color_bottom = self._fetch_color_from_position(0)
            color_top = self._fetch_color_from_position(1)

            if color_bottom is not None and color_top is not None:
                if color_bottom == color_top:
                    # Both are the same, show that color
                    self._update_ui_from_rgb(color_bottom)
                else:
                    # Different colors, show black
                    self._update_ui_from_rgb((0, 0, 0))
            else:
                # Could not fetch one or both colors, show black
                self._update_ui_from_rgb((0, 0, 0))
        else:  # Bottom (0) or Top (1)
            color = self._fetch_color_from_position(position)
            if color is not None:
                self._update_ui_from_rgb(color)
            else:
                # Could not fetch color, show black
                self._update_ui_from_rgb((0, 0, 0))

    def _fetch_color_from_position(self, position):
        """
        Fetch current color from Crazyflie for given position

        Args:
            position: 0 for bottom, 1 for top

        Returns:
            tuple (r, g, b) or None if fetch failed
        """
        if position not in self.PARAMS_BY_POSITION:
            return None

        if not self._deck_controller.is_deck_present(position):
            return None

        try:
            param_name = self.PARAMS_BY_POSITION[position]['color']
            color_uint32 = int(self._helper.cf.param.get_value(param_name))

            # Unpack WRGB: 0xWWRRGGBB
            w = (color_uint32 >> 24) & 0xFF
            r = (color_uint32 >> 16) & 0xFF
            g = (color_uint32 >> 8) & 0xFF
            b = color_uint32 & 0xFF

            # Convert WRGB back to RGB by adding white channel back
            r_full = r + w
            g_full = g + w
            b_full = b + w

            return (r_full, g_full, b_full)
        except (KeyError, ValueError, TypeError) as e:
            logger.debug(f"Could not fetch color from position {position}: {e}")
            return None

    def _update_ui_from_rgb(self, rgb):
        """
        Update UI controls from RGB values without writing back to Crazyflie

        Args:
            rgb: tuple (r, g, b)
        """
        self._updating_from_fetch = True
        try:
            r, g, b = rgb
            color = QColor(r, g, b)
            h, s, v, _ = color.getHsvF()

            self._hue = h or 0
            self._saturation = s or 0
            self._value = v or 0

            self.hue_bar.setValue(int(self._hue * 1000))
            self._update_sv_area(self.sv_area, self._hue)
            self._update_preview()
        finally:
            self._updating_from_fetch = False

    def _populate_position_dropdown(self):
        """Initialize the position dropdown with items (called once during __init__)"""
        self.positionDropdown.addItem("Bottom", 0)
        self.positionDropdown.addItem("Top", 1)
        self.positionDropdown.addItem("Both", 2)

        # Initially all disabled until we connect and detect decks
        self.positionDropdown.model().item(0).setEnabled(False)
        self.positionDropdown.model().item(1).setEnabled(False)
        self.positionDropdown.model().item(2).setEnabled(False)

    def _update_position_dropdown(self):
        """Update position dropdown based on detected decks"""
        is_bottom_attached = self._deck_controller.is_deck_present(0)
        is_top_attached = self._deck_controller.is_deck_present(1)

        if is_bottom_attached and is_top_attached:
            self.positionDropdown.model().item(0).setEnabled(True)
            self.positionDropdown.model().item(1).setEnabled(True)
            self.positionDropdown.model().item(2).setEnabled(True)
            # Default to "Both" if both attached
            self.positionDropdown.setCurrentIndex(2)
        elif is_bottom_attached:
            self.positionDropdown.model().item(0).setEnabled(True)
            self.positionDropdown.model().item(1).setEnabled(False)
            self.positionDropdown.model().item(2).setEnabled(False)
            self.positionDropdown.setCurrentIndex(0)
        elif is_top_attached:
            self.positionDropdown.model().item(0).setEnabled(False)
            self.positionDropdown.model().item(1).setEnabled(True)
            self.positionDropdown.model().item(2).setEnabled(False)
            self.positionDropdown.setCurrentIndex(1)
        else:
            # No decks attached
            self.positionDropdown.model().item(0).setEnabled(False)
            self.positionDropdown.model().item(1).setEnabled(False)
            self.positionDropdown.model().item(2).setEnabled(False)

    def showEvent(self, a0):
        """ Show event for proper initial SV area sizing """
        super().showEvent(a0)
        self._update_sv_area(self.sv_area, self._hue)
        # Update preview without writing to Crazyflie
        self._updating_from_fetch = True
        try:
            self._update_preview()
        finally:
            self._updating_from_fetch = False

    def mousePressEvent(self, a0):
        self._handle_mouse_event(a0)

    def mouseMoveEvent(self, a0):
        self._handle_mouse_event(a0)

    def _handle_mouse_event(self, event):
        if not self.groupBox_color.isEnabled():
            return
        sv_pos = self.sv_area.mapFrom(self, event.pos())
        if self.sv_area.rect().contains(sv_pos):
            self._update_sv_from_pos(sv_pos)

    def _update_sv_from_pos(self, pos):
        x = pos.x()
        y = pos.y()
        w = self.sv_area.width()
        h = self.sv_area.height()

        self._saturation = max(0, min(1, x / w))
        self._value = 1 - max(0, min(1, y / h))
        self._update_preview()
        self._update_sv_area(self.sv_area, self._hue)

    def _update_sv_area(self, sv_area, hue):
        width = sv_area.width()
        height = sv_area.height()
        if width <= 0 or height <= 0:
            return

        pixmap = QPixmap(width, height)
        pixmap.fill(Qt.GlobalColor.transparent)

        painter = QPainter(pixmap)
        painter.setRenderHint(QPainter.RenderHint.Antialiasing)

        """ Rounded corners """
        path = QPainterPath()
        corner_radius = 8
        path.addRoundedRect(0, 0, width, height, corner_radius, corner_radius)
        painter.setClipPath(path)

        """ Base hue layer """
        base_color = QColor.fromHsvF(hue, 1, 1)
        painter.fillRect(pixmap.rect(), base_color)

        """ Saturation overlay (white → transparent) """
        sat_gradient = QLinearGradient(0, 0, width, 0)
        sat_gradient.setColorAt(0, Qt.GlobalColor.white)
        sat_gradient.setColorAt(1, Qt.GlobalColor.transparent)
        painter.fillRect(pixmap.rect(), sat_gradient)

        """ Value overlay (transparent → black) """
        val_gradient = QLinearGradient(0, 0, 0, height)
        val_gradient.setColorAt(0, Qt.GlobalColor.transparent)
        val_gradient.setColorAt(1, Qt.GlobalColor.black)
        painter.fillRect(pixmap.rect(), val_gradient)

        """ Outline """
        outline_pen = QPen(Qt.GlobalColor.darkGray)
        outline_pen.setWidth(1)
        painter.setPen(outline_pen)
        painter.setBrush(Qt.BrushStyle.NoBrush)
        painter.drawPath(path)

        """ Circle selector """
        sel_x = int(self._saturation * width)
        sel_y = int((1 - self._value) * height)
        selector_radius = 6
        pen = QPen(Qt.GlobalColor.white)
        pen.setWidth(2)
        painter.setPen(pen)
        painter.setBrush(Qt.BrushStyle.NoBrush)
        painter.drawEllipse(sel_x - selector_radius, sel_y - selector_radius, selector_radius * 2, selector_radius * 2)

        painter.end()
        sv_area.setPixmap(pixmap)

    def _on_hue_slider_changed(self, value):
        if not self.groupBox_color.isEnabled():
            return
        self._hue = value / 1000
        self._update_sv_area(self.sv_area, self._hue)
        self._update_preview()

    def _update_preview(self):
        color = QColor.fromHsvF(self._hue or 0, self._saturation or 0, self._value or 0)
        self.color_preview.setStyleSheet(f"""
        QFrame#color_preview {{
            border: 1px solid #444;
            border-radius: 4px;
        }}

        QFrame#color_preview:enabled {{
            background-color: {color.name()};
        }}

        QFrame#color_preview:disabled {{
            background-color: #777777;  /* greyed out when disconnected */
        }}
        """)
        self.hex_input.setText(color.name().upper())
        self._colorChanged.emit(color)

    def _on_hex_changed(self):
        if not self.groupBox_color.isEnabled():
            return
        hex_value = self.hex_input.text().strip()
        color = QColor(hex_value)
        if color.isValid():
            h, s, v, _ = color.getHsvF()
            self._hue, self._saturation, self._value = h or 0, s or 0, v or 0
            self.hue_bar.setValue(int(self._hue * 1000))
            self._update_sv_area(self.sv_area, self._hue)
            self._update_preview()
            self.hex_input.setStyleSheet("")
            self.hex_error_label.setText("")
            self.information_text.setText("")
        else:
            self.hex_input.setStyleSheet("border: 2px solid red; border-radius: 4px;")
            self.hex_error_label.setText("Invalid hex code.")
            self.information_text.setText("")
            logger.warning(f"Invalid HEX color: {hex_value}")

    def _connect_color_buttons(self):
        color_buttons = [
            self.color_button1,
            self.color_button2,
            self.color_button3,
            self.color_button4,
            self.color_button5,
            self.color_button6,
            self.color_button7,
            self.color_button8,
            self.color_button9,
            self.color_button10,
        ]
        for btn in color_buttons:
            btn.clicked.connect(self._on_color_button_clicked)

    def _on_color_button_clicked(self):
        if not self.groupBox_color.isEnabled():
            return
        button = self.sender()
        style = button.styleSheet()  # type: ignore
        if "background-color:" not in style:
            return
        hex_color = style.split("background-color:")[-1].split(";")[0].strip()
        color = QColor(hex_color)
        if color.isValid():
            h, s, v, _ = color.getHsvF()
            self._hue, self._saturation, self._value = h or 0, s or 0, v or 0
            self.hue_bar.setValue(int(self._hue * 1000))
            self._update_sv_area(self.sv_area, self._hue)
            self._update_preview()

    def _add_custom_color_button(self):
        color_hex = self.hex_input.text().strip()
        if not color_hex.startswith("#"):
            color_hex = "#" + color_hex

        color = QColor(color_hex)
        if not color.isValid():
            logger.warning(f"Invalid custom color: {color_hex}")
            return

        new_btn = QPushButton()
        new_btn.setStyleSheet(f"background-color: {color_hex};")
        new_btn.setFixedSize(50, 30)
        new_btn.clicked.connect(self._on_color_button_clicked)
        new_btn.setContextMenuPolicy(Qt.ContextMenuPolicy.CustomContextMenu)
        new_btn.customContextMenuRequested.connect(lambda pos, btn=new_btn: self._remove_custom_color_button(btn))

        self.custom_color_buttons.append(new_btn)

        self._repack_custom_buttons()
        logger.debug(f"Added new custom color {color_hex}")

    def _remove_custom_color_button(self, button):
        reply = QMessageBox.question(
            self,
            "Remove Custom Color",
            "Do you want to remove this custom color?",
            QMessageBox.StandardButton.Yes | QMessageBox.StandardButton.No,
        )
        if reply == QMessageBox.StandardButton.Yes:
            if button in self.custom_color_buttons:
                self.custom_color_buttons.remove(button)
            button.setParent(None)
            button.deleteLater()
            logger.debug("Removed custom color button.")

            self._repack_custom_buttons()

    def _repack_custom_buttons(self):
        grid = self.gridLayout_5
        plus_button = self.add_color_button

        total_cols = 8
        row, col = 0, 0

        preset_buttons = [
            self.color_button1,
            self.color_button2,
            self.color_button3,
            self.color_button4,
            self.color_button5,
            self.color_button6,
            self.color_button7,
            self.color_button8,
            self.color_button9,
            self.color_button10,
        ]

        for btn in self.custom_color_buttons + [plus_button]:
            grid.removeWidget(btn)

        for btn in preset_buttons:
            grid.addWidget(btn, row, col)
            col += 1
            if col >= total_cols:
                col = 0
                row += 1

        for btn in self.custom_color_buttons:
            grid.addWidget(btn, row, col)
            col += 1
            if col >= total_cols:
                col = 0
                row += 1

        grid.addWidget(plus_button, row, col)



# -*- coding: utf-8 -*-
#     ||
#  +------+      / __ )(_) /_______________ _____  ___
#  | 0xBC |     / __  / / __/ ___/ ___/ __ `/_  / / _ \
#  +------+    / /_/ / / /_/ /__/ /  / /_/ / / /_/  __/
#   ||  ||    /_____/_/\__/\___/_/   \__,_/ /___/\___/
#
#  Copyright (C) 2011-2013 Bitcraze AB
#
#  Crazyflie Nano Quadcopter Client
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either version 2
#  of the License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software Foundation,
#  Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
"""
Linux joystick driver using the Linux input_joystick subsystem. Requires sysfs
to be mounted on /sys and /dev/input/js* to be readable.

This module is very linux specific but should work on any CPU platform
"""
import ctypes
import glob
import logging
import os
import struct
import sys

if not sys.platform.startswith('linux'):
    raise Exception("Only supported on Linux")

try:
    import fcntl
except ImportError as e:
    raise Exception("fcntl library probably not installed ({})".format(e))

__author__ = 'Bitcraze AB'
__all__ = ['Joystick']

logger = logging.getLogger(__name__)

JS_EVENT_FMT = "@IhBB"
JE_TIME = 0
JE_VALUE = 1
JE_TYPE = 2
JE_NUMBER = 3

JS_EVENT_BUTTON = 0x001
JS_EVENT_AXIS = 0x002
JS_EVENT_INIT = 0x080

# ioctls
JSIOCGAXES = 0x80016a11
JSIOCGBUTTONS = 0x80016a12

MODULE_MAIN = "Joystick"
MODULE_NAME = "linuxjsdev"


class JEvent(object):
    """
    Joystick event class. Encapsulate single joystick event.
    """

    def __init__(self, evt_type, number, value):
        self.type = evt_type
        self.number = number
        self.value = value

    def __repr__(self):
        return "JEvent(type={}, number={}, value={})".format(
            self.type, self.number, self.value)


# Constants
TYPE_BUTTON = 1
TYPE_AXIS = 2


class _JS():

    def __init__(self, num, name):
        self.num = num
        self.name = name
        self._f_name = "/dev/input/js{}".format(num)
        self._f = None

        self.opened = False
        self.buttons = []
        self.axes = []
        self._prev_pressed = {}

    def open(self):
        if self._f:
            raise Exception("{} at {} is already "
                            "opened".format(self.name, self._f_name))

        self._f = open("/dev/input/js{}".format(self.num), "rb")
        fcntl.fcntl(self._f.fileno(), fcntl.F_SETFL, os.O_NONBLOCK)

        # Get number of axis and button
        val = ctypes.c_int()
        if fcntl.ioctl(self._f.fileno(), JSIOCGAXES, val) != 0:
            self._f.close()
            self._f = None
            raise Exception("Failed to read number of axes")

        self.axes = list(0 for i in range(val.value))
        if fcntl.ioctl(self._f.fileno(), JSIOCGBUTTONS, val) != 0:
            self._f.close()
            self._f = None
            raise Exception("Failed to read number of axes")

        self.buttons = list(0 for i in range(val.value))
        self.__initvalues()

    def close(self):
        """Open the joystick device"""
        if not self._f:
            return

        logger.info("Closed {} ({})".format(self.name, self.num))

        self._f.close()
        self._f = None

    def __initvalues(self):
        """Read the buttons and axes initial values from the js device"""
        for _ in range(len(self.axes) + len(self.buttons)):
            data = self._f.read(struct.calcsize(JS_EVENT_FMT))
            jsdata = struct.unpack(JS_EVENT_FMT, data)
            self.__updatestate(jsdata)

    def __updatestate(self, jsdata):
        """Update the internal absolute state of buttons and axes"""
        if jsdata[JE_TYPE] & JS_EVENT_AXIS != 0:
            self.axes[jsdata[JE_NUMBER]] = jsdata[JE_VALUE] / 32768.0
        elif jsdata[JE_TYPE] & JS_EVENT_BUTTON != 0:
            self.buttons[jsdata[JE_NUMBER]] = jsdata[JE_VALUE]

    def __decode_event(self, jsdata):
        """ Decode a jsdev event into a dict """
        # TODO: Add timestamp?
        if jsdata[JE_TYPE] & JS_EVENT_AXIS != 0:
            return JEvent(evt_type=TYPE_AXIS,
                          number=jsdata[JE_NUMBER],
                          value=jsdata[JE_VALUE] / 32768.0)
        if jsdata[JE_TYPE] & JS_EVENT_BUTTON != 0:
            return JEvent(evt_type=TYPE_BUTTON,
                          number=jsdata[JE_NUMBER],
                          value=jsdata[JE_VALUE] / 32768.0)

    def _read_all_events(self):
        """Consume all the events queued up in the JS device"""
        try:
            while True:
                data = self._f.read(struct.calcsize(JS_EVENT_FMT))
                jsdata = struct.unpack(JS_EVENT_FMT, data)
                self.__updatestate(jsdata)
        except IOError as e:
            if e.errno != 11:
                logger.info(str(e))
                self._f.close()
                self._f = None
                raise IOError("Device has been disconnected")
        except TypeError:
            pass
        except ValueError:
            # This will happen if I/O operations are done on a closed device,
            # which is the case when you first close and then open the device
            # while switching device. But, in order for SDL2 to work on Linux
            # (for debugging) the device needs to be closed before it's opened.
            # This is the workaround to make both cases work.
            pass

    def read(self):
        """ Returns a list of all joystick event since the last call """
        if not self._f:
            raise Exception("Joystick device not opened")

        self._read_all_events()

        return [self.axes, self.buttons]


class Joystick():
    """
    Linux jsdev implementation of the Joystick class
    """

    def __init__(self):
        self.name = MODULE_NAME
        self._js = {}
        self._devices = []

    def devices(self):
        """
        Returns a dict with device_id as key and device name as value of all
        the detected devices (result is cached once one or more device are
        found).
        """

        if len(self._devices) == 0:
            syspaths = glob.glob("/sys/class/input/js*")

            for path in syspaths:
                device_id = int(os.path.basename(path)[2:])
                with open(path + "/device/name") as namefile:
                    name = namefile.read().strip()
                self._js[device_id] = _JS(device_id, name)
                self._devices.append({"id": device_id, "name": name})

        return self._devices

    def open(self, device_id):
        """
        Open the joystick device. The device_id is given by available_devices
        """
        self._js[device_id].open()

    def close(self, device_id):
        """Open the joystick device"""
        self._js[device_id].close()

    def read(self, device_id):
        """ Returns a list of all joystick event since the last call """
        return self._js[device_id].read()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
#     ||          ____  _ __
#  +------+      / __ )(_) /_______________ _____  ___
#  | 0xBC |     / __  / / __/ ___/ ___/ __ `/_  / / _ \
#  +------+    / /_/ / / /_/ /__/ /  / /_/ / / /_/  __/
#   ||  ||    /_____/_/\__/\___/_/   \__,_/ /___/\___/
#
#  Copyright (C) 2021-2023 Bitcraze AB
#
#  Crazyflie Nano Quadcopter Client
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either version 2
#  of the License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.

#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
#  02110-1301, USA.
from collections import namedtuple
from PyQt6.QtCore import pyqtSignal, QObject

__author__ = 'Bitcraze AB'
__all__ = ['ConnectivityManager']


class ConnectivityManager(QObject):
    UiElementsContainer = namedtuple('UiElementContainer', [
        'interface_combo',
        'address_spinner',
        'connect_button',
        'scan_button'])

    class UIState:
        DISCONNECTED = 0
        CONNECTING = 1
        CONNECTED = 2
        SCANNING = 3

    INTERFACE_PROMPT_TEXT = 'Select an interface'

    connect_button_clicked = pyqtSignal()
    scan_button_clicked = pyqtSignal(object)
    connection_state_changed = pyqtSignal(object)

    def __init__(self):
        QObject.__init__(self)
        self._ui_elements = []
        self._state = self.UIState.DISCONNECTED
        self._is_enabled = True

    def register_ui_elements(self, ui_elements):
        self._ui_elements.append(ui_elements)

        ui_elements.connect_button.clicked.connect(self._connect_button_click_handler)
        ui_elements.scan_button.clicked.connect(self._scan_button_click_handler)

        ui_elements.address_spinner.valueChanged.connect(self._address_changed_handler)
        ui_elements.address_spinner.editingFinished.connect(self._address_edited_handler)

        ui_elements.interface_combo.currentIndexChanged.connect(self._interface_combo_current_index_changed_handler)

    def set_state(self, state):
        if self._state != state:
            self._state = state
            self._update_ui()

            if self._state == self.UIState.DISCONNECTED:
                self.connection_state_changed.emit(self.UIState.DISCONNECTED)
            elif self._state == self.UIState.CONNECTED:
                self.connection_state_changed.emit(self.UIState.CONNECTED)
            elif self._state == self.UIState.CONNECTING:
                self.connection_state_changed.emit(self.UIState.CONNECTING)
            elif self._state == self.UIState.SCANNING:
                self.connection_state_changed.emit(self.UIState.SCANNING)

    def set_enable(self, enable):
        if self._is_enabled != enable:
            self._is_enabled = enable
            self._update_ui()

    def set_address(self, address):
        for ui_elements in self._ui_elements:
            ui_elements.address_spinner.setValue(address)

    def get_address(self):
        if len(self._ui_elements) > 0:
            return self._ui_elements[0].address_spinner.value()
        else:
            return 0

    def set_interfaces(self, interface_items, index):
        new_index = 0
        if index is not None:
            new_index = index + 1

        for ui_elements in self._ui_elements:
            combo = ui_elements.interface_combo

            combo.clear()
            combo.addItem(self.INTERFACE_PROMPT_TEXT)
            combo.addItems(interface_items)
            combo.setCurrentIndex(new_index)

    def get_interface(self):
        if len(self._ui_elements) > 0:
            interface = self._ui_elements[0].interface_combo.currentText()
            if interface == self.INTERFACE_PROMPT_TEXT:
                self._selected_interface = None
            else:
                return interface
        else:
            return None

    def _connect_button_click_handler(self):
        self.connect_button_clicked.emit()

    def _scan_button_click_handler(self):
        self.scan_button_clicked.emit(self.get_address())

    def _address_changed_handler(self, value):
        for ui_elements in self._ui_elements:
            if value != ui_elements.address_spinner.value():
                ui_elements.address_spinner.setValue(value)

    def _address_edited_handler(self):
        # Find out if one of the addresses has changed and what the new value is
        value = 0
        is_changed = False
        for ui_elements in self._ui_elements:
            if ui_elements.address_spinner.is_text_different_from_value():
                value = ui_elements.address_spinner.value()
                is_changed = True
                break

        # Set the new value
        if is_changed:
            for ui_elements in self._ui_elements:
                if value != ui_elements.address_spinner.value():
                    ui_elements.address_spinner.setValue(value)

    def _interface_combo_current_index_changed_handler(self, interface):
        interface_s = str(interface)
        can_connect = interface != self.INTERFACE_PROMPT_TEXT
        for ui_elements in self._ui_elements:
            combo = ui_elements.interface_combo
            if combo.currentText != interface_s:
                combo.setCurrentText(interface_s)
            ui_elements.connect_button.setEnabled(can_connect)

    def _update_ui(self):
        if self._is_enabled:
            if self._state == self.UIState.DISCONNECTED:
                can_connect = self.get_interface() is not None
                for ui_elements in self._ui_elements:
                    ui_elements.connect_button.setText("Connect")
                    ui_elements.connect_button.setToolTip("Connect to the Crazyflie on the selected interface (Ctrl+I)")
                    ui_elements.connect_button.setEnabled(can_connect)
                    ui_elements.scan_button.setText("Scan")
                    ui_elements.scan_button.setEnabled(True)
                    ui_elements.address_spinner.setEnabled(True)
                    ui_elements.interface_combo.setEnabled(True)
            elif self._state == self.UIState.CONNECTED:
                for ui_elements in self._ui_elements:
                    ui_elements.connect_button.setText("Disconnect")
                    ui_elements.connect_button.setToolTip("Disconnect from the Crazyflie (Ctrl+I)")
                    ui_elements.scan_button.setEnabled(False)
                    ui_elements.address_spinner.setEnabled(False)
                    ui_elements.interface_combo.setEnabled(False)
            elif self._state == self.UIState.CONNECTING:
                for ui_elements in self._ui_elements:
                    ui_elements.connect_button.setText("Cancel")
                    ui_elements.connect_button.setToolTip("Cancel connecting to the Crazyflie")
                    ui_elements.scan_button.setEnabled(False)
                    ui_elements.address_spinner.setEnabled(False)
                    ui_elements.interface_combo.setEnabled(False)
            elif self._state == self.UIState.SCANNING:
                for ui_elements in self._ui_elements:
                    ui_elements.connect_button.setText("Connect")
                    ui_elements.connect_button.setEnabled(False)
                    ui_elements.scan_button.setText("Scanning...")
                    ui_elements.scan_button.setEnabled(False)
                    ui_elements.address_spinner.setEnabled(False)
                    ui_elements.interface_combo.setEnabled(False)
        else:
            for ui_elements in self._ui_elements:
                ui_elements.connect_button.setEnabled(False)
                ui_elements.scan_button.setEnabled(False)
                ui_elements.address_spinner.setEnabled(False)
                ui_elements.interface_combo.setEnabled(False)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
#     ||          ____  _ __
#  +------+      / __ )(_) /_______________ _____  ___
#  | 0xBC |     / __  / / __/ ___/ ___/ __ `/_  / / _ \
#  +------+    / /_/ / / /_/ /__/ /  / /_/ / / /_/  __/
#   ||  ||    /_____/_/\__/\___/_/   \__,_/ /___/\___/
#
#  Copyright (C) 2011-2023 Bitcraze AB
#
#  Crazyflie Nano Quadcopter Client
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either version 2
#  of the License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.

#  You should have received a copy of the GNU General Public License along with
#  this program; if not, write to the Free Software Foundation, Inc.,
#  51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

"""
The console tab is used as a console for printouts from the Crazyflie.
"""

import logging

from PyQt6 import uic
from PyQt6.QtCore import pyqtSignal
from PyQt6.QtGui import QTextCursor

import cfclient
from cfclient.ui.tab_toolbox import TabToolbox

__author__ = 'Bitcraze AB'
__all__ = ['ConsoleTab']

logger = logging.getLogger(__name__)

console_tab_class = uic.loadUiType(cfclient.module_path +
                                   "/ui/tabs/consoleTab.ui")[0]


class ConsoleTab(TabToolbox, console_tab_class):
    """Console tab for showing printouts from Crazyflie"""
    _link_established_signal = pyqtSignal(str)
    _connected_signal = pyqtSignal(str)
    _disconnected_signal = pyqtSignal(str)
    _update = pyqtSignal(str)

    def __init__(self, helper):
        super(ConsoleTab, self).__init__(helper, 'Console')
        self.setupUi(self)

        # Always wrap callbacks from Crazyflie API though QT Signal/Slots
        # to avoid manipulating the UI when rendering it
        self._link_established_signal.connect(self._link_established)
        self._connected_signal.connect(self._connected)
        self._disconnected_signal.connect(self._disconnected)
        self._update.connect(self.printText)

        self._helper.cf.console.receivedChar.add_callback(self._update.emit)
        self._helper.cf.connected.add_callback(self._connected_signal.emit)
        self._helper.cf.link_established.add_callback(self._link_established_signal.emit)
        self._helper.cf.disconnected.add_callback(
            self._disconnected_signal.emit)

        self._clearButton.clicked.connect(self.clear)
        self._dumpSystemLoadButton.clicked.connect(
            lambda enabled:
            self._helper.cf.param.set_value("system.taskDump", '1'))
        self._dumpAssertInformation.clicked.connect(
            lambda enabled:
            self._helper.cf.param.set_value_raw("system.assertInfo", 0x08, 1))
        self._propellerTestButton.clicked.connect(
            lambda enabled:
            self._helper.cf.param.set_value("health.startPropTest", '1'))
        self._batteryTestButton.clicked.connect(
            lambda enabled:
            self._helper.cf.param.set_value("health.startBatTest", '1'))
        self._storageStatsButton.clicked.connect(
            lambda enabled:
            self._helper.cf.param.set_value("system.storageStats", '1'))

    def printText(self, text):
        # Make sure we get printouts from the Crazyflie into the log (such as
        # build version and test ok/fail)
        logger.debug("[%s]", text)
        scrollbar = self.console.verticalScrollBar()
        prev_scroll = scrollbar.value()
        prev_cursor = self.console.textCursor()
        was_maximum = prev_scroll == scrollbar.maximum()

        self.console.moveCursor(QTextCursor.MoveOperation.End)
        self.console.insertPlainText(text)

        self.console.setTextCursor(prev_cursor)

        if was_maximum and not prev_cursor.hasSelection():
            scrollbar.setValue(scrollbar.maximum())
        else:
            scrollbar.setValue(prev_scroll)

    def clear(self):
        self.console.clear()

    def _connected(self, link_uri):
        """Callback when the Crazyflie has been connected"""
        self._dumpSystemLoadButton.setEnabled(True)
        self._propellerTestButton.setEnabled(True)
        self._batteryTestButton.setEnabled(True)
        self._storageStatsButton.setEnabled(True)

    def _disconnected(self, link_uri):
        """Callback for when the Crazyflie has been disconnected"""
        self._dumpSystemLoadButton.setEnabled(False)
        self._dumpAssertInformation.setEnabled(False)
        self._propellerTestButton.setEnabled(False)
        self._batteryTestButton.setEnabled(False)
        self._storageStatsButton.setEnabled(False)

    def _link_established(self, link_uri):
        """Callback when the first packet on a new link is received"""
        # Enable the assert dump button as early as possible. After an assert we will never get the connected() cb.
        self._dumpAssertInformation.setEnabled(True)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
#     ||          ____  _ __
#  +------+      / __ )(_) /_______________ _____  ___
#  | 0xBC |     / __  / / __/ ___/ ___/ __ `/_  / / _ \
#  +------+    / /_/ / / /_/ /__/ /  / /_/ / / /_/  __/
#   ||  ||    /_____/_/\__/\___/_/   \__,_/ /___/\___/
#
#  Copyright (C) 2015 Bitcraze AB
#
#  Crazyflie Nano Quadcopter Client
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either version 2
#  of the License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.

#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301,
#  USA.

"""
Test script to show how to use the ZMQ LED-ring backend for the Crazyflie
Python Client. The backend is started automatically in the client if ZMQ
is available.

The example will fill up the ring by sending 12 message, each with a new
LED lit up, until all of them are lit up in the last message.
"""

try:
    import zmq
except ImportError as e:
    raise Exception("ZMQ library probably not installed ({})".format(e))

zmess = {
    "version": 1,
    "rgbleds": [
        [0, 0, 0],
        [0, 0, 0],
        [0, 0, 0],
        [0, 0, 0],
        [0, 0, 0],
        [0, 0, 0],
        [0, 0, 0],
        [0, 0, 0],
        [0, 0, 0],
        [0, 0, 0],
        [0, 0, 0],
        [0, 0, 0]
    ]
}

context = zmq.Context()
receiver = context.socket(zmq.PUSH)
bind_addr = "tcp://127.0.0.1:{}".format(1024 + 190)
receiver.connect(bind_addr)

for i in range(12):
    zmess["rgbleds"][i] = [128, 0, 128]
    receiver.send_json(zmess)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
#     ||          ____  _ __
#  +------+      / __ )(_) /_______________ _____  ___
#  | 0xBC |     / __  / / __/ ___/ ___/ __ `/_  / / _ \
#  +------+    / /_/ / / /_/ /__/ /  / /_/ / / /_/  __/
#   ||  ||    /_____/_/\__/\___/_/   \__,_/ /___/\___/
#
#  Copyright (C) 2014 Bitcraze AB
#
#  Crazyflie Nano Quadcopter Client
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either version 2
#  of the License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
#  02110-1301, USA.
"""
Mux for giving control to one device (slave/student) for all axis (roll/pitch/
yaw/thrust) with the ability to take over all of them from a second device
(master/teacher).
"""
import logging

from .takeoverselectivemux import TakeOverSelectiveMux

__author__ = 'Bitcraze AB'
__all__ = ['TakeOverMux']

logger = logging.getLogger(__name__)


class TakeOverMux(TakeOverSelectiveMux):

    def __init__(self, *args):
        super(TakeOverMux, self).__init__(*args)
        self.name = "Teacher (RPYT)"
        self._muxing = {
            self._master: ("estop", "alt1", "alt2", "assistedControl", "exit"),
            self._slave: ("roll", "pitch", "yaw", "thrust")
        }



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
#     ||          ____  _ __
#  +------+      / __ )(_) /_______________ _____  ___
#  | 0xBC |     / __  / / __/ ___/ ___/ __ `/_  / / _ \
#  +------+    / /_/ / / /_/ /__/ /  / /_/ / / /_/  __/
#   ||  ||    /_____/_/\__/\___/_/   \__,_/ /___/\___/
#
#  Copyright (C) 2011-2024 Bitcraze AB
#
#  Crazyflie Nano Quadcopter Client
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either version 2
#  of the License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.

#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA  02110-1301, USA.

"""
The logconfigreader module is responsible for reading and parsing log
configuration data. This data is used to control what information is logged
in the client.
"""

import json
import logging
import os
import re
import shutil

import cfclient
from cflib.crazyflie.log import LogVariable, LogConfig

from PyQt6 import QtGui

__author__ = 'Bitcraze AB'
__all__ = ['LogVariable', 'LogConfigReader']

logger = logging.getLogger(__name__)

DEFAULT_CONF_NAME = 'log_config'
DEFAULT_CATEGORY_NAME = 'category'

FILE_REGEX_YAML = "Config *.yaml;;All *.*"


class LogConfigReader():
    """Reads logging configurations from file"""

    def __init__(self, crazyflie):

        self._log_configs = {}
        self.dsList = []
        # Check if user config exists, otherwise copy files
        if (not os.path.exists(cfclient.config_path + "/log")):
            logger.info("No user config found, copying dist files")
            shutil.copytree(cfclient.module_path + "/configs/log", cfclient.config_path + "/log")
        self._cf = crazyflie
        self._cf.connected.add_callback(self._connected)

    def get_icons(self):
        client_path = os.path.abspath(os.path.join(os.path.dirname(__file__),
                                      os.pardir))
        icon_path = os.path.join(client_path, 'ui', 'icons')
        save_icon = QtGui.QIcon(os.path.join(icon_path, 'create.png'))
        delete_icon = QtGui.QIcon(os.path.join(icon_path, 'delete.png'))
        return save_icon, delete_icon

    def create_empty_log_conf(self, category):
        """ Creates an empty log-configuration with a default name """
        log_path = self._get_log_path(category)
        conf_name = self._get_default_conf_name(log_path)
        file_path = os.path.join(log_path, conf_name) + '.json'

        if not os.path.exists(file_path):
            with open(file_path, 'w') as f:
                f.write(json.dumps(
                    {
                        'logconfig': {
                            'logblock': {
                                'variables': [],
                                'name': conf_name,
                                'period': 100
                            }
                        }
                    }, indent=2))

        self._log_configs[category].append(LogConfig(conf_name, 100))
        return conf_name

    def create_category(self):
        """ Creates a new category (dir in filesystem), with a unique name """
        log_path = os.path.join(cfclient.config_path, 'log')
        category = self._get_default_category(log_path)
        dir_path = os.path.join(log_path, category)

        # This should never be false, but just to be safe.
        if not os.path.exists(dir_path):
            os.mkdir(dir_path)
            self._log_configs[category] = []

        return category

    def delete_category(self, category):
        """ Removes the directory on file-system and recursively removes
            all the logging configurations.
        """
        log_path = self._get_log_path(category)
        if os.path.exists(log_path):
            shutil.rmtree(log_path)
            self._log_configs.pop(category)

    def delete_config(self, conf_name, category):
        """ Deletes a configuration from file system. """
        log_path = self._get_log_path(category)
        conf_path = os.path.join(log_path, conf_name) + '.json'

        if not os.path.exists(conf_path):
            # Check if we can find the file with lowercase first letter.
            conf_path = os.path.join(log_path,
                                     conf_name[0].lower() + conf_name[1:]
                                     + '.json')
            if not os.path.exists(conf_path):
                # Cant' find the config-file
                logger.warning('Failed to find log-config %s' % conf_path)
                return

        os.remove(conf_path)
        for conf in self._log_configs[category]:
            if conf.name == conf_name:
                self._log_configs[category].remove(conf)

    def change_name_config(self, old_name, new_name, category):
        """ Changes name to the configuration and updates the
            file in the file system.
        """
        configs = self._log_configs[category]

        for conf in configs:
            if conf.name == old_name:
                conf.name = new_name

        log_path = self._get_log_path(category)
        old_path = os.path.join(log_path, old_name) + '.json'
        new_path = os.path.join(log_path, new_name) + '.json'

        # File should exist but just to be extra safe
        if os.path.exists(old_path):
            with open(old_path, 'r+') as f:
                data = json.load(f)
                data['logconfig']['logblock']['name'] = new_name
                f.seek(0)
                f.truncate()
                f.write(json.dumps(data, indent=2))

            os.rename(old_path, new_path)

    def change_name_category(self, old_name, new_name):
        """ Renames the directory on file system and the config dict """
        if old_name in self._log_configs:
            self._log_configs[new_name] = self._log_configs.pop(old_name)
            os.rename(self._get_log_path(old_name),
                      self._get_log_path(new_name))

    def _get_log_path(self, category):
        """ Helper method """
        category_dir = '' if category == 'Default' else '/' + category
        return os.path.join(cfclient.config_path,
                            'log' + category_dir)

    def _get_default_category(self, log_path):
        """ Creates a name for the category, ending with a unique number. """
        dirs = [dir_ for dir_ in os.listdir(log_path) if os.path.isdir(
            os.path.join(log_path, dir_)
        )]
        config_nbrs = re.findall(r'(?<=%s)\d*' % DEFAULT_CATEGORY_NAME,
                                 ' '.join(dirs))
        config_nbrs = list(filter(len, config_nbrs))

        if config_nbrs:
            return DEFAULT_CATEGORY_NAME + str(
                            max([int(nbr) for nbr in config_nbrs]) + 1)
        else:
            return DEFAULT_CATEGORY_NAME + '1'

    def _read_config_categories(self):
        """Read and parse log configurations"""

        self._log_configs = {'Default': []}
        log_path = os.path.join(cfclient.config_path, 'log')

        for category in os.listdir(log_path):

            category_path = os.path.join(log_path, category)

            try:
                if (os.path.isdir(category_path)):
                    # create a new cathegory
                    self._log_configs[category] = []
                    for conf in os.listdir(category_path):
                        if conf.endswith('.json'):
                            conf_path = os.path.join(category_path, conf)
                            log_conf = self._get_conf(conf_path)

                        # add the log configuration to the cathegory
                        self._log_configs[category].append(log_conf)

                else:
                    # if it's not a directory, the log config is placed
                    # in the 'Default' cathegory
                    if category_path.endswith('.json'):
                        log_conf = self._get_conf(category_path)
                        self._log_configs['Default'].append(log_conf)

            except Exception as e:
                logger.warning("Failed to open log config %s", e)

    def _get_default_conf_name(self, log_path):
        config_nbrs = re.findall(r'(?<=%s)\d*(?!=\.json)' % DEFAULT_CONF_NAME,
                                 ' '.join(os.listdir(log_path)))
        config_nbrs = list(filter(len, config_nbrs))

        if config_nbrs:
            return DEFAULT_CONF_NAME + str(
                        max([int(nbr) for nbr in config_nbrs]) + 1)
        else:
            return DEFAULT_CONF_NAME + '1'

    def _get_conf(self, conf_path):
        with open(conf_path) as f:
            data = json.load(f)
            infoNode = data["logconfig"]["logblock"]

            logConf = LogConfig(infoNode["name"],
                                int(infoNode["period"]))
            for v in data["logconfig"]["logblock"]["variables"]:
                if v["type"] == "TOC":
                    logConf.add_variable(str(v["name"]), v["fetch_as"])
                else:
                    logConf.add_variable("Mem", v["fetch_as"],
                                         v["stored_as"],
                                         int(v["address"], 16))
            return logConf

    def _get_configpaths_recursively(self):
        """ Reads all configuration files from the log path and
            returns a list of tuples with format:
            (category/conf-name, absolute path).
        """
        logpath = os.path.join(cfclient.config_path, 'log')
        filepaths = []

        for files in os.listdir(logpath):
            abspath = os.path.join(logpath, files)
            if os.path.isdir(abspath):
                for config in os.listdir(abspath):
                    if config.endswith('.json'):
                        filepaths.append(('/'.join([files, config]),
                                         os.path.join(abspath, config)))
            else:
                if files.endswith('.json'):
                    filepaths.append((files, os.path.join(abspath)))

        return filepaths

    def _read_config_files(self):
        """Read and parse log configurations"""

        configsfound = self._get_configpaths_recursively()

        new_dsList = []
        for conf in configsfound:
            try:
                logger.info("Parsing [%s]", conf[0])
                json_data = open(conf[1])
                self.data = json.load(json_data)
                infoNode = self.data["logconfig"]["logblock"]
                logConfName = conf[0].replace('.json', '')

                logConf = LogConfig(logConfName, int(infoNode["period"]))
                for v in self.data["logconfig"]["logblock"]["variables"]:
                    if v["type"] == "TOC":
                        logConf.add_variable(str(v["name"]), v["fetch_as"])
                    else:
                        logConf.add_variable("Mem", v["fetch_as"],
                                             v["stored_as"],
                                             int(v["address"], 16))
                new_dsList.append(logConf)
                json_data.close()
            except Exception as e:
                logger.warning("Exception while parsing logconfig file: %s", e)
        self.dsList = new_dsList

    def _connected(self, link_uri):
        """Callback that is called once Crazyflie is connected"""

        self._read_config_files()
        self._read_config_categories()
        # Just add all the configurations. Via callbacks other parts of the
        # application will pick up these configurations and use them
        for d in self.dsList:
            try:
                self._cf.log.add_config(d)
            except KeyError as e:
                logger.warning(str(e))
            except AttributeError as e:
                logger.warning(str(e))

    def getLogConfigs(self):
        """Return the log configurations"""
        return self.dsList

    def _getLogConfigs(self):
        """Return the log configurations"""
        return self._log_configs

    def saveLogConfigFile(self, category, logconfig):
        """Save a log configuration to file"""
        log_path = self._get_log_path(category)
        file_path = os.path.join(log_path, logconfig.name) + '.json'
        logger.info("Saving config for [%s]", file_path)

        # Build tree for JSON
        saveConfig = {}
        logconf = {'logblock': {'variables': []}}
        logconf['logblock']['name'] = logconfig.name
        logconf['logblock']['period'] = logconfig.period_in_ms
        # Temporary until plot is fixed

        for v in logconfig.variables:
            newC = {}
            newC['name'] = v.name
            newC['stored_as'] = v.stored_as_string
            newC['fetch_as'] = v.fetch_as_string
            newC['type'] = "TOC"
            logconf['logblock']['variables'].append(newC)

        saveConfig['logconfig'] = logconf

        for old_conf in self._log_configs[category]:
            if old_conf.name == logconfig.name:
                self._log_configs[category].remove(old_conf)
                self._log_configs[category].append(logconfig)

        with open(file_path, 'w') as f:
            f.write(json.dumps(saveConfig, indent=2))

        self._read_config_files()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
#     ||          ____  _ __
#  +------+      / __ )(_) /_______________ _____  ___
#  | 0xBC |     / __  / / __/ ___/ ___/ __ `/_  / / _ \
#  +------+    / /_/ / / /_/ /__/ /  / /_/ / / /_/  __/
#   ||  ||    /_____/_/\__/\___/_/   \__,_/ /___/\___/
#
#  Copyright (C) 2022-2023 Bitcraze AB
#
#  Crazyflie Nano Quadcopter Client
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either version 2
#  of the License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.

#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
#  02110-1301, USA.

"""
Shows data for the Lighthouse Positioning system
"""

import logging

from PyQt6 import uic
from PyQt6.QtCore import Qt, pyqtSignal, QTimer
from PyQt6.QtWidgets import QMessageBox
from PyQt6.QtWidgets import QFileDialog
from PyQt6.QtWidgets import QLabel

import cfclient
from cfclient.ui.tab_toolbox import TabToolbox

from cflib.crazyflie.log import LogConfig
from cflib.crazyflie.mem import LighthouseMemHelper
from cflib.localization import LighthouseConfigWriter
from cflib.localization import LighthouseConfigFileManager

from cfclient.ui.dialogs.lighthouse_bs_geometry_dialog import LighthouseBsGeometryDialog
from cfclient.ui.dialogs.basestation_mode_dialog import LighthouseBsModeDialog
from cfclient.ui.dialogs.lighthouse_system_type_dialog import LighthouseSystemTypeDialog
from cfclient.utils.logconfigreader import FILE_REGEX_YAML

from vispy import scene
import numpy as np
import math
import os

__author__ = 'Bitcraze AB'
__all__ = ['LighthouseTab']

logger = logging.getLogger(__name__)

lighthouse_tab_class = uic.loadUiType(cfclient.module_path + "/ui/tabs/lighthouse_tab.ui")[0]

STYLE_RED_BACKGROUND = "background-color: lightpink;"
STYLE_GREEN_BACKGROUND = "background-color: lightgreen;"
STYLE_BLUE_BACKGROUND = "background-color: lightblue;"
STYLE_ORANGE_BACKGROUND = "background-color: orange;"
STYLE_NO_BACKGROUND = "background-color: none;"


class MarkerPose():
    COL_X_AXIS = 'red'
    COL_Y_AXIS = 'green'
    COL_Z_AXIS = 'blue'

    AXIS_LEN = 0.3

    LABEL_SIZE = 100
    LABEL_OFFSET = np.array((0.0, 0, 0.25))

    def __init__(self, the_scene, color, text=None):
        self._scene = the_scene
        self._color = color
        self._text = text

        self._marker = scene.visuals.Markers(
            pos=np.array([[0, 0, 0]]),
            parent=self._scene,
            face_color=self._color)

        self._x_axis = scene.visuals.Line(
            pos=np.array([[0, 0, 0], [0, 0, 0]]),
            color=self.COL_X_AXIS,
            parent=self._scene)

        self._y_axis = scene.visuals.Line(pos=np.array(
            [[0, 0, 0], [0, 0, 0]]),
            color=self.COL_Y_AXIS,
            parent=self._scene)

        self._z_axis = scene.visuals.Line(
            pos=np.array([[0, 0, 0], [0, 0, 0]]),
            color=self.COL_Z_AXIS,
            parent=self._scene)

        self._label = None
        if self._text:
            self._label = scene.visuals.Text(
                text=self._text,
                font_size=self.LABEL_SIZE,
                pos=self.LABEL_OFFSET,
                parent=self._scene)

    def set_pose(self, position, rot):
        self._marker.set_data(pos=np.array([position]), face_color=self._color)

        if self._label:
            self._label.pos = self.LABEL_OFFSET + position

        x_tip = np.dot(np.array(rot), np.array([self.AXIS_LEN, 0, 0]))
        self._x_axis.set_data(np.array([position, x_tip + position]), color=self.COL_X_AXIS)

        y_tip = np.dot(np.array(rot), np.array([0, self.AXIS_LEN, 0]))
        self._y_axis.set_data(np.array([position, y_tip + position]), color=self.COL_Y_AXIS)

        z_tip = np.dot(np.array(rot), np.array([0, 0, self.AXIS_LEN]))
        self._z_axis.set_data(np.array([position, z_tip + position]), color=self.COL_Z_AXIS)

    def remove(self):
        self._marker.parent = None
        self._x_axis.parent = None
        self._y_axis.parent = None
        self._z_axis.parent = None
        if self._label:
            self._label.parent = None

    def set_color(self, color):
        self._color = color
        self._marker.set_data(face_color=self._color)


class Plot3dLighthouse(scene.SceneCanvas):
    POSITION_BRUSH = np.array((0, 0, 1.0))
    BS_BRUSH_VISIBLE = np.array((0.2, 0.5, 0.2))
    BS_BRUSH_NOT_VISIBLE = np.array((0.8, 0.5, 0.5))

    VICINITY_DISTANCE = 2.5
    HIGHLIGHT_DISTANCE = 0.5

    LABEL_SIZE = 100
    LABEL_HIGHLIGHT_SIZE = 200

    HIGHLIGHT_SIZE = 20

    TEXT_OFFSET = np.array((0.0, 0, 0.25))

    def __init__(self):
        scene.SceneCanvas.__init__(self, keys=None)
        self.unfreeze()

        self._view = self.central_widget.add_view()
        self._view.bgcolor = '#ffffff'
        self._view.camera = scene.TurntableCamera(
            distance=10.0,
            up='+z',
            center=(0.0, 0.0, 1.0))

        self._cf = None
        self._base_stations = {}

        self.freeze()

        plane_size = 10
        scene.visuals.Plane(
            width=plane_size,
            height=plane_size,
            width_segments=plane_size,
            height_segments=plane_size,
            color=(0.5, 0.5, 0.5, 0.5),
            edge_color="gray",
            parent=self._view.scene)

        self._addArrows(1, 0.02, 0.1, 0.1, self._view.scene)

    def _addArrows(self, length, width, head_length, head_width, parent):
        # The Arrow visual in vispy does not seem to work very good,
        # draw arrows using lines instead.
        w = width / 2
        hw = head_width / 2
        base_len = length - head_length

        # X-axis
        scene.visuals.LinePlot([
            [0, w, 0],
            [base_len, w, 0],
            [base_len, hw, 0],
            [length, 0, 0],
            [base_len, -hw, 0],
            [base_len, -w, 0],
            [0, -w, 0]],
            width=1.0, color='red', parent=parent, marker_size=0.0)

        # Y-axis
        scene.visuals.LinePlot([
            [w, 0, 0],
            [w, base_len, 0],
            [hw, base_len, 0],
            [0, length, 0],
            [-hw, base_len, 0],
            [-w, base_len, 0],
            [-w, 0, 0]],
            width=1.0, color='green', parent=parent, marker_size=0.0)

        # Z-axis
        scene.visuals.LinePlot([
            [0, w, 0],
            [0, w, base_len],
            [0, hw, base_len],
            [0, 0, length],
            [0, -hw, base_len],
            [0, -w, base_len],
            [0, -w, 0]],
            width=1.0, color='blue', parent=parent, marker_size=0.0)

    def update_cf_pose(self, position, rot):
        if not self._cf:
            self._cf = MarkerPose(self._view.scene, self.POSITION_BRUSH)
        self._cf.set_pose(position, rot)

    def update_base_station_geos(self, geos):
        for id, geo in geos.items():
            if (geo is not None) and (id not in self._base_stations):
                self._base_stations[id] = MarkerPose(self._view.scene, self.BS_BRUSH_NOT_VISIBLE, text=f"{id + 1}")
            self._base_stations[id].set_pose(geo.origin, geo.rotation_matrix)

        geos_to_remove = self._base_stations.keys() - geos.keys()
        for id in geos_to_remove:
            existing = self._base_stations.pop(id)
            existing.remove()

    def update_base_station_visibility(self, visibility):
        for id, bs in self._base_stations.items():
            if id in visibility:
                bs.set_color(self.BS_BRUSH_VISIBLE)
            else:
                bs.set_color(self.BS_BRUSH_NOT_VISIBLE)

    def clear(self):
        if self._cf:
            self._cf.remove()
            self._cf = None

        for bs in self._base_stations.values():
            bs.remove()
        self._base_stations = {}

    def _mix(self, col1, col2, mix):
        return col1 * mix + col2 * (1.0 - mix)


class LighthouseTab(TabToolbox, lighthouse_tab_class):
    """Tab for plotting Lighthouse data"""

    # Update period of log data in ms
    UPDATE_PERIOD_LOG = 100

    # Frame rate (updates per second)
    FPS = 2

    STATUS_NOT_RECEIVING = 0
    STATUS_MISSING_DATA = 1
    STATUS_TO_ESTIMATOR = 2

    # TODO change these names to something more logical
    LOG_STATUS = "lighthouse.status"
    LOG_RECEIVE = "lighthouse.bsReceive"
    LOG_CALIBRATION_EXISTS = "lighthouse.bsCalVal"
    LOG_CALIBRATION_CONFIRMED = "lighthouse.bsCalCon"
    LOG_CALIBRATION_UPDATED = "lighthouse.bsCalUd"
    LOG_GEOMETERY_EXISTS = "lighthouse.bsGeoVal"
    LOG_ACTIVE = "lighthouse.bsActive"
    LOG_AVAILABLE = "lighthouse.bsAvailable"

    _connected_signal = pyqtSignal(str)
    _disconnected_signal = pyqtSignal(str)
    _log_error_signal = pyqtSignal(object, str)
    _status_report_signal = pyqtSignal(int, object, object)
    _new_system_config_written_to_cf_signal = pyqtSignal(bool)
    _geometry_read_signal = pyqtSignal(object)
    _calibration_read_signal = pyqtSignal(object)

    def __init__(self, helper):
        super(LighthouseTab, self).__init__(helper, 'Lighthouse Positioning')
        self.setupUi(self)

        # Always wrap callbacks from Crazyflie API though QT Signal/Slots
        # to avoid manipulating the UI when rendering it
        self._connected_signal.connect(self._connected)
        self._disconnected_signal.connect(self._disconnected)
        self._log_error_signal.connect(self._logging_error)
        self._status_report_signal.connect(self._status_report_received)
        self._new_system_config_written_to_cf_signal.connect(self._new_system_config_written_to_cf)
        self._geometry_read_signal.connect(self._geometry_read_cb)
        self._calibration_read_signal.connect(self._calibration_read_cb)

        # Connect the Crazyflie API callbacks to the signals
        self._helper.cf.connected.add_callback(self._connected_signal.emit)
        self._helper.cf.disconnected.add_callback(self._disconnected_signal.emit)

        self._set_up_plots()

        self.is_lighthouse_deck_active = False

        self._lh_memory_helper = None
        self._lh_config_writer = None
        self._lh_geos = {}
        self._is_geometry_read_ongoing = False

        self._bs_receives_light = set()
        self._bs_calibration_data_exists = set()
        self._bs_calibration_data_confirmed = set()
        self._bs_calibration_data_updated = set()
        self._bs_geometry_data_exists = set()
        self._bs_data_to_estimator = set()
        self._bs_available = set()

        self._clear_state_indicator()

        self._bs_stats = [
            self._bs_receives_light,
            self._bs_calibration_data_exists,
            self._bs_calibration_data_confirmed,
            self._bs_calibration_data_updated,
            self._bs_geometry_data_exists,
            self._bs_data_to_estimator,
            self._bs_available]

        self._lh_status = self.STATUS_NOT_RECEIVING

        self._graph_timer = QTimer()
        self._graph_timer.setInterval(int(1000 / self.FPS))
        self._graph_timer.timeout.connect(self._update_graphics)
        self._graph_timer.start()

        self._basestation_geometry_dialog = LighthouseBsGeometryDialog(self)
        self._basestation_mode_dialog = LighthouseBsModeDialog(self)
        self._system_type_dialog = LighthouseSystemTypeDialog(helper)

        self._manage_estimate_geometry_button.clicked.connect(self._show_basestation_geometry_dialog)
        self._change_system_type_button.clicked.connect(lambda: self._system_type_dialog.show())
        self._manage_basestation_mode_button.clicked.connect(self._show_basestation_mode_dialog)

        self._load_sys_config_button.clicked.connect(self._load_sys_config_button_clicked)
        self._save_sys_config_button.clicked.connect(self._save_sys_config_button_clicked)

        self._is_connected = False
        self._update_ui()

    def write_and_store_geometry(self, geometries):
        if self._lh_config_writer:
            self._lh_config_writer.write_and_store_config(self._new_system_config_written_to_cf_signal.emit,
                                                          geos=geometries)

    def _new_system_config_written_to_cf(self, success):
        # Reset the bit fields for calibration data status to get a fresh view
        self._helper.cf.param.set_value("lighthouse.bsCalibReset", '1')
        # New geo data has been written and stored in the CF, read it back to update the UI
        self._start_read_of_geo_data()

    def _show_basestation_geometry_dialog(self):
        self._basestation_geometry_dialog.reset()
        self._basestation_geometry_dialog.show()

    def _show_basestation_mode_dialog(self):
        self._basestation_mode_dialog.reset()
        self._basestation_mode_dialog.show()

    def _set_up_plots(self):
        self._plot_3d = Plot3dLighthouse()
        self._plot_layout.addWidget(self._plot_3d.native)

    def _connected(self, link_uri):
        """Callback when the Crazyflie has been connected"""
        logger.debug("Crazyflie connected to {}".format(link_uri))

        self._basestation_geometry_dialog.reset()
        self._is_connected = True

        if self._helper.cf.param.get_value('deck.bcLighthouse4') == '1':
            self._lighthouse_deck_detected()

        self._update_ui()

    def _lighthouse_deck_detected(self):
        """Called when the lighthouse deck has been detected. Enables the tab,
        starts logging and polling of the memory sub system as well as starts
        timers for updating graphics"""
        if not self.is_lighthouse_deck_active:
            self.is_lighthouse_deck_active = True

            try:
                self._register_logblock(
                    "lhStatus",
                    [self.LOG_STATUS, self.LOG_RECEIVE, self.LOG_CALIBRATION_EXISTS, self.LOG_CALIBRATION_CONFIRMED,
                        self.LOG_CALIBRATION_UPDATED, self.LOG_GEOMETERY_EXISTS, self.LOG_ACTIVE, self.LOG_AVAILABLE],
                    self._status_report_signal.emit,
                    self._log_error_signal.emit)
            except KeyError as e:
                logger.warning(str(e))
            except AttributeError as e:
                logger.warning(str(e))

            self._populate_status_matrix()

            # Now that we know we have a lighthouse deck, setup the memory helper and config writer
            self._lh_memory_helper = LighthouseMemHelper(self._helper.cf)
            self._lh_config_writer = LighthouseConfigWriter(self._helper.cf)

    def _start_read_of_geo_data(self):
        if not self._is_geometry_read_ongoing:
            self._is_geometry_read_ongoing = True
            self._lh_memory_helper.read_all_geos(self._geometry_read_signal.emit)

    def _geometry_read_cb(self, geometries):
        # Remove any geo data where the valid flag is False
        self._lh_geos = dict(filter(lambda key_value: key_value[1].valid, geometries.items()))
        self._basestation_geometry_dialog.geometry_updated(self._lh_geos)
        self._is_geometry_read_ongoing = False

    def _is_matching_current_geo_data(self, geometries):
        return geometries == self._lh_geos.keys()

    def _adjust_bitmask(self, bit_mask, bs_list):
        for id in range(16):
            if bit_mask & (1 << id):
                bs_list.add(id)
            else:
                if id in bs_list:
                    bs_list.remove(id)

    def _status_report_received(self, timestamp, data, logconf):
        """Callback from the logging system when the status is updated."""

        if self.LOG_RECEIVE in data:
            bit_mask = data[self.LOG_RECEIVE]
            self._adjust_bitmask(bit_mask, self._bs_receives_light)
        if self.LOG_CALIBRATION_EXISTS in data:
            bit_mask = data[self.LOG_CALIBRATION_EXISTS]
            self._adjust_bitmask(bit_mask, self._bs_calibration_data_exists)
        if self.LOG_CALIBRATION_CONFIRMED in data:
            bit_mask = data[self.LOG_CALIBRATION_CONFIRMED]
            self._adjust_bitmask(bit_mask, self._bs_calibration_data_confirmed)
        if self.LOG_CALIBRATION_UPDATED in data:
            bit_mask = data[self.LOG_CALIBRATION_UPDATED]
            self._adjust_bitmask(bit_mask, self._bs_calibration_data_updated)
        if self.LOG_GEOMETERY_EXISTS in data:
            bit_mask = data[self.LOG_GEOMETERY_EXISTS]
            self._adjust_bitmask(bit_mask, self._bs_geometry_data_exists)
            if not self._is_matching_current_geo_data(self._bs_geometry_data_exists):
                self._start_read_of_geo_data()

        if self.LOG_ACTIVE in data:
            bit_mask = data[self.LOG_ACTIVE]
            self._adjust_bitmask(bit_mask, self._bs_data_to_estimator)

        if self.LOG_STATUS in data:
            self._lh_status = data[self.LOG_STATUS]

        if self.LOG_AVAILABLE in data:
            bit_mask = data[self.LOG_AVAILABLE]
            self._adjust_bitmask(bit_mask, self._bs_available)

        self._update_basestation_status_indicators()

    def _disconnected(self, link_uri):
        """Callback for when the Crazyflie has been disconnected"""
        logger.debug("Crazyflie disconnected from {}".format(link_uri))
        self._clear_state()
        self._update_graphics()
        self._plot_3d.clear()
        self._basestation_geometry_dialog.close()
        self.is_lighthouse_deck_active = False
        self._is_connected = False
        self._update_ui()

    def _register_logblock(self, logblock_name, variables, data_cb, error_cb,
                           update_period=UPDATE_PERIOD_LOG):
        """Register log data to listen for. One logblock can only contain a limited
        number of parameters."""
        lg = LogConfig(logblock_name, update_period)
        for variable in variables:
            if self._is_in_log_toc(variable):
                lg.add_variable(variable)

        self._helper.cf.log.add_config(lg)
        lg.data_received_cb.add_callback(data_cb)
        lg.error_cb.add_callback(error_cb)
        lg.start()
        return lg

    def _is_in_log_toc(self, variable):
        toc = self._helper.cf.log.toc
        group, param = variable.split('.')
        return group in toc.toc and param in toc.toc[group]

    def _is_in_param_toc(self, group, param):
        toc = self._helper.cf.param.toc
        return bool(group in toc.toc and param in toc.toc[group])

    def _logging_error(self, log_conf, msg):
        """Callback from the log layer when an error occurs"""
        QMessageBox.about(self, "LighthouseTab error",
                          "Error when using log config",
                          " [{0}]: {1}".format(log_conf.name, msg))

    def _update_graphics(self):
        if self.is_visible() and self.is_lighthouse_deck_active:
            self._plot_3d.update_cf_pose(self._helper.pose_logger.position,
                                         self._rpy_to_rot(self._helper.pose_logger.rpy_rad))
            self._plot_3d.update_base_station_geos(self._lh_geos)
            self._plot_3d.update_base_station_visibility(self._bs_data_to_estimator)
            self._update_position_label(self._helper.pose_logger.position)
            self._update_status_label(self._lh_status)
            self._mask_status_matrix(self._bs_available)

    def _update_ui(self):
        enabled = self._is_connected and self.is_lighthouse_deck_active
        self._manage_estimate_geometry_button.setEnabled(enabled)
        self._change_system_type_button.setEnabled(enabled)
        self._load_sys_config_button.setEnabled(enabled)
        self._save_sys_config_button.setEnabled(enabled)

    def _update_position_label(self, position):
        if len(position) == 3:
            coordinate = "({:0.2f}, {:0.2f}, {:0.2f})".format(
                position[0], position[1], position[2])
        else:
            coordinate = '(0.00, 0.00, 0.00)'

        self._status_position.setText(coordinate)

    def _update_status_label(self, status):
        text = ''
        if status == self.STATUS_NOT_RECEIVING:
            text = 'Not receiving'
        elif status == self.STATUS_MISSING_DATA:
            text = 'No geo/calib'
        elif status == self.STATUS_TO_ESTIMATOR:
            text = 'LH ready'

        self._status_status.setText(text)

    def _clear_state(self):
        self._lh_memory_helper = None
        self._lh_config_writer = None
        self._lh_geos = {}
        self._is_geometry_read_ongoing = False
        self._bs_receives_light.clear()
        self._bs_calibration_data_exists.clear()
        self._bs_calibration_data_confirmed.clear()
        self._bs_calibration_data_updated.clear()
        self._bs_geometry_data_exists.clear()
        self._bs_data_to_estimator.clear()
        self._update_basestation_status_indicators()
        self._clear_state_indicator()
        self._lh_status = self.STATUS_NOT_RECEIVING

    def _clear_state_indicator(self):
        container = self._basestation_stats_container
        for row in range(0, 5):
            for col in range(1, 17):
                item = container.itemAtPosition(row, col)
                if item is not None:
                    item.widget().deleteLater()

    def _rpy_to_rot(self, rpy):
        # http://planning.cs.uiuc.edu/node102.html
        # Pitch reversed compared to page above
        roll = rpy[0]
        pitch = rpy[1]
        yaw = rpy[2]

        cg = math.cos(roll)
        cb = math.cos(-pitch)
        ca = math.cos(yaw)
        sg = math.sin(roll)
        sb = math.sin(-pitch)
        sa = math.sin(yaw)

        r = [
            [ca * cb, ca * sb * sg - sa * cg, ca * sb * cg + sa * sg],
            [sa * cb, sa * sb * sg + ca * cg, sa * sb * cg - ca * sg],
            [-sb, cb * sg, cb * cg],
        ]

        return np.array(r)

    def _populate_status_matrix(self):
        container = self._basestation_stats_container

        # Find the nr of base stations by looking for the highest bit that is set
        # Assume all bs up to that bit are available
        for bs in range(0, 16):
            container.addWidget(self._create_label(str(bs + 1)), 0, bs + 1)
            for i in range(1, 5):
                container.addWidget(self._create_label(), i, bs + 1)

    def _mask_status_matrix(self, bs_available_mask):
        container = self._basestation_stats_container

        # Find the nr of base stations by looking for the highest bit that is set
        # Assume all bs up to that bit are available
        for bs in range(0, 16):
            bs_indicator_id = bs + 1
            for stats_indicator_id in range(0, 5):
                item = container.itemAtPosition(stats_indicator_id, bs_indicator_id)
                if item is not None:
                    label = item.widget()
                    if bs_indicator_id - 1 in bs_available_mask:
                        label.setHidden(False)
                    else:
                        label.setHidden(True)

    def _create_label(self, text=None):
        label = QLabel()
        label.setMinimumSize(30, 0)
        label.setAlignment(Qt.AlignmentFlag.AlignCenter)

        if text:
            label.setText(str(text))
        else:
            label.setProperty('frameShape', 'QFrame::Box')
            label.setStyleSheet(STYLE_NO_BACKGROUND)

        return label

    def _update_basestation_status_indicators(self):
        """Handling the basestation status label handles to indicate
            the state of received data per basestation"""
        container = self._basestation_stats_container

        # Ports the label number to the first index of the statistic id
        stats_id_port = {1: 0, 2: 1, 3: 4, 4: 5}

        for bs in range(16):
            for stats_indicator_id in range(1, 5):
                bs_indicator_id = bs + 1
                item = container.itemAtPosition(stats_indicator_id, bs_indicator_id)
                if item is not None:
                    label = item.widget()
                    stats_id = stats_id_port.get(stats_indicator_id)
                    temp_set = self._bs_stats[stats_id]

                    if bs in temp_set:
                        # If the status bar for calibration data is handled, have an intermediate status
                        # else just have red or green.
                        if stats_indicator_id == 2:
                            label.setStyleSheet(STYLE_BLUE_BACKGROUND)
                            label.setToolTip('Calibration data from cache')

                            calib_confirm = bs in self._bs_stats[stats_id + 1]
                            calib_updated = bs in self._bs_stats[stats_id + 2]

                            if calib_confirm:
                                label.setStyleSheet(STYLE_GREEN_BACKGROUND)
                                label.setToolTip('Calibration data verified')
                            if calib_updated:
                                label.setStyleSheet(STYLE_ORANGE_BACKGROUND)
                                label.setToolTip('Calibration data updated, the geometry probably needs to be ' +
                                                 're-estimated')
                        else:
                            label.setStyleSheet(STYLE_GREEN_BACKGROUND)
                    else:
                        label.setStyleSheet(STYLE_RED_BACKGROUND)
                        label.setToolTip('')

    def _load_sys_config_button_clicked(self):
        names = QFileDialog.getOpenFileName(self, 'Open file', self._helper.current_folder, FILE_REGEX_YAML)

        if names[0] == '':
            return

        self._helper.current_folder = os.path.dirname(names[0])

        if self._lh_config_writer is not None:
            self._lh_config_writer.write_and_store_config_from_file(self._new_system_config_written_to_cf_signal.emit,
                                                                    names[0])

    def _save_sys_config_button_clicked(self):
        # Get calibration data from the Crazyflie to complete the system config data set
        # When the data is ready we get a callback on _calibration_read
        self._lh_memory_helper.read_all_calibs(self._calibration_read_signal.emit)

    def _calibration_read_cb(self, calibs):
        # Got calibration data from the CF, we have the full system configuration
        system_type = self._system_type_dialog.get_system_type()
        self._save_sys_config(self._lh_geos, calibs, system_type)

    def _save_sys_config(self, geos, calibs, system_type):
        names = QFileDialog.getSaveFileName(self, 'Save file', self._helper.current_folder, FILE_REGEX_YAML)

        if names[0] == '':
            return

        self._helper.current_folder = os.path.dirname(names[0])

        if not names[0].endswith(".yaml") and names[0].find(".") < 0:
            filename = names[0] + ".yaml"
        else:
            filename = names[0]

        LighthouseConfigFileManager.write(filename, geos=geos, calibs=calibs, system_type=system_type)



# -*- coding: utf-8 -*-
#
# ,---------,       ____  _ __
# |  ,-^-,  |      / __ )(_) /_______________ _____  ___
# | (  O  ) |     / __  / / __/ ___/ ___/ __ `/_  / / _ \
# | / ,--'  |    / /_/ / / /_/ /__/ /  / /_/ / / /_/  __/
#    +------`   /_____/_/\__/\___/_/   \__,_/ /___/\___/
#
# Copyright (C) 2022-2023 Bitcraze AB
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, in version 3.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""
Slider widget with advanced features
"""

from PyQt6 import QtWidgets, QtCore
from cflib.utils.callbacks import Caller


__author__ = 'Bitcraze AB'
__all__ = ['SuperSlider']


class SuperSlider(QtWidgets.QWidget):

    def __init__(self, min: float, max: float, value: float):
        super(SuperSlider, self).__init__()

        self.slider_scaling: int = 100

        self.min = min
        self.max = max
        self.value = value

        self.value_changed_cb = Caller()
        self._initUI()

    def set_value(self, value: float):
        self.input.blockSignals(True)
        self.slider.blockSignals(True)

        self.value = value
        self.input.setValue(self.value)
        self.slider.setValue(int(self.value * self.slider_scaling))

        self.input.blockSignals(False)
        self.slider.blockSignals(False)

    def _initUI(self):
        # Create controls
        self.slider = QtWidgets.QSlider(QtCore.Qt.Orientation.Horizontal, self)
        self.slider.setFocusPolicy(QtCore.Qt.FocusPolicy.NoFocus)
        self.slider.setRange(int(self.min * self.slider_scaling), int(self.max * self.slider_scaling))
        self.slider.setValue(int(self.value * self.slider_scaling))
        self.slider.setTickInterval(self.slider_scaling)
        self.slider.setTickPosition(QtWidgets.QSlider.TickPosition.TicksBelow)

        self.min_label = QtWidgets.QLabel(str(self.min), self)
        self.min_label.setAlignment(QtCore.Qt.AlignmentFlag.AlignLeft)
        self._small_font(self.min_label)

        self.input = QtWidgets.QDoubleSpinBox(self)
        self.input.setMinimum(self.min)
        self.input.setMaximum(self.max)
        self.input.setSingleStep(0.1)
        self.input.setValue(self.value)

        self.max_label = QtWidgets.QLabel(str(self.max), self)
        self.max_label.setAlignment(QtCore.Qt.AlignmentFlag.AlignRight)
        self._small_font(self.max_label)

        # Connect functionality
        self.slider.valueChanged.connect(self._slider_moved)
        self.input.valueChanged.connect(self._input_changed)

        # Layout
        layout = QtWidgets.QGridLayout()
        layout.addWidget(self.slider, 0, 0, 1, 3)
        layout.addWidget(self.min_label, 1, 0)
        layout.addWidget(self.input, 1, 1)
        layout.addWidget(self.max_label, 1, 2)
        self.setLayout(layout)

    def _small_font(self, label):
        font = label.font()
        font.setPointSize(int(font.pointSize() * 0.8))
        label.setFont(font)

    def _slider_moved(self, value):
        self.input.blockSignals(True)
        self.value = value / self.slider_scaling
        self.input.setValue(self.value)
        self.input.blockSignals(False)
        self.value_changed_cb.call(self.value)

    def _input_changed(self, value):
        self.slider.blockSignals(True)
        self.value = value
        self.slider.setValue(int(self.value * self.slider_scaling))
        self.slider.blockSignals(False)
        self.value_changed_cb.call(self.value)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
#     ||          ____  _ __
#  +------+      / __ )(_) /_______________ _____  ___
#  | 0xBC |     / __  / / __/ ___/ ___/ __ `/_  / / _ \
#  +------+    / /_/ / / /_/ /__/ /  / /_/ / / /_/  __/
#   ||  ||    /_____/_/\__/\___/_/   \__,_/ /___/\___/
#
#  Copyright (C) 2015 Bitcraze AB
#
#  Crazyflie Nano Quadcopter Client
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either version 2
#  of the License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.

#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301,
#  USA.

"""
Give access to the parameter framework via ZMQ.
"""

import logging
from threading import Thread, Lock

from cfclient.utils.config import Config

ZMQ_PULL_PORT = 1024 + 189
logger = logging.getLogger(__name__)

enabled = False
try:
    import zmq

    enabled = True
except Exception as e:
    logger.warning(
        "Not enabling ZMQ param access, import failed ({})".format(e))

if not Config().get("enable_zmq_param"):
    enabled = False
    logger.info("ZMQ param disabled in config file")


class _PullReader(Thread):

    def __init__(self, receiver, callback, *args):
        super(_PullReader, self).__init__(*args)
        self._receiver = receiver
        self._cb = callback
        self.daemon = True
        self.lock = Lock()

    def run(self):
        while True:
            self.lock.acquire()
            self._cb(self._receiver.recv_json())


class ZMQParamAccess:
    """Used for reading data from input devices using the PyGame API."""

    def __init__(self, crazyflie):

        if enabled:
            self._cf = crazyflie
            context = zmq.Context()
            self._receiver = context.socket(zmq.REP)
            self._bind_addr = "tcp://*:{}".format(ZMQ_PULL_PORT)
            # If the port is already bound an exception will be thrown
            # and caught in the initialization of the readers and handled.
            self._receiver.bind(self._bind_addr)
            logger.info(
                "Biding ZMQ for parameters at {}".format(self._bind_addr))
            self._receiver_thread = _PullReader(self._receiver,
                                                self._cmd_callback)

    def start(self):
        if enabled:
            self._receiver_thread.start()

    def _cmd_callback(self, data):
        # logger.info(data)
        if data["cmd"] == "toc":
            response = {"version": 1, "toc": []}
            self._receiver.send_json(response)
            self._receiver_thread.lock.release()
        if data["cmd"] == "set":
            resp = {"version": 1}  # noqa
            group = data["name"].split(".")[0]
            name = data["name"].split(".")[1]
            self._cf.param.add_update_callback(group=group, name=name,
                                               cb=self._param_callback)
            self._cf.param.set_value(data["name"], str(data["value"]))

    def _param_callback(self, name, value):
        group = name.split(".")[0]
        name_short = name.split(".")[1]
        logger.info("Removing {}.{}".format(group, name_short))
        self._cf.param.remove_update_callback(group=group, name=name_short,
                                              cb=self._param_callback)

        response = {"version": 1, "cmd": "set", "name": name, "value": value}
        self._receiver.send_json(response)
        self._receiver_thread.lock.release()
